"""Generate Golang constants corresponding to the constants of the meta-model."""
import io
from typing import (
    Optional,
    List,
    Tuple,
)

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.golang import (
    common as golang_common,
    naming as golang_naming,
    description as golang_description,
)
from aas_core_codegen.golang.common import (
    INDENT as I,
)


# region Generation


def _generate_documentation_comment_for_constant(
    description: intermediate.DescriptionOfConstant,
    context: golang_description.Context,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given constant."""
    # fmt: off
    comment, errors = (
        golang_description
        .generate_comment_for_summary_remarks(
            description=description, context=context
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None
    return comment, None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_primitive(
    constant: intermediate.ConstantPrimitive,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant primitive."""
    writer = io.StringIO()

    if constant.description is not None:
        comment, comment_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=golang_description.Context(
                package=golang_common.CONSTANTS_PACKAGE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            return None, Error(
                constant.parsed.node,
                f"Failed to generate the documentation comment for {constant.name!r}",
                comment_errors,
            )

        assert comment is not None
        writer.write(comment)
        writer.write("\n")

    constant_name = golang_naming.constant_name(constant.name)

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        assert isinstance(constant.value, bool)
        literal = golang_common.boolean_literal(constant.value)
        writer.write(f"const {constant_name} bool = {literal}")

    elif constant.a_type is intermediate.PrimitiveType.INT:
        assert isinstance(constant.value, int)

        if constant.value > 2**63 - 1 or constant.value < -(2**63):
            return None, Error(
                constant.parsed.node,
                f"The value of the constant {constant.name!r} overflows "
                f"64-bit signed integer",
            )
        writer.write(f"const {constant_name} int64 = {str(constant.value)}")

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        assert isinstance(constant.value, float)

        # NOTE (mristin, 2023-03-29):
        # We assume that the float constants are not meant to be all to precise.
        # Therefore, we use a string representation here. However, beware that we
        # might have to use a more precise representation in the future if the spec
        # change.
        literal = golang_common.float_literal(constant.value)

        writer.write(f"const {constant_name} float64 = {literal}")

    elif constant.a_type is intermediate.PrimitiveType.STR:
        assert isinstance(constant.value, str)
        literal = golang_common.string_literal(constant.value)

        writer.write(f"const {constant_name} string = {literal}")

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        assert isinstance(constant.value, bytearray)

        literal, _ = golang_common.bytes_literal(value=constant.value)

        writer.write(
            f"""\
var {constant_name} = {indent_but_first_line(literal, I)}"""
        )

    else:
        assert_never(constant.a_type)

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_primitives(
    constant: intermediate.ConstantSetOfPrimitives,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of primitives."""
    errors = []  # type: List[Error]

    writer = io.StringIO()

    if constant.description is not None:
        comment, comment_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=golang_description.Context(
                package=golang_common.CONSTANTS_PACKAGE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            errors.append(
                Error(
                    constant.parsed.node,
                    f"Failed to generate the documentation comment for {constant.name!r}",
                    comment_errors,
                )
            )
        else:
            assert comment is not None
            writer.write(comment)
            writer.write("\n")

    literal_codes = []  # type: List[str]
    set_type: str

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        set_type = "map[bool]struct{}"
        for literal in constant.literals:
            assert isinstance(literal.value, bool)
            literal_codes.append(golang_common.boolean_literal(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.INT:
        set_type = "map[int64]struct{}"
        for literal in constant.literals:
            assert isinstance(literal.value, int)

            literal_codes.append(str(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        set_type = "map[float64]struct{}"

        for literal in constant.literals:
            assert isinstance(literal.value, float)

            literal_codes.append(golang_common.float_literal(literal.value))

    elif constant.a_type is intermediate.PrimitiveType.STR:
        set_type = "map[string]struct{}"

        for literal in constant.literals:
            assert isinstance(literal.value, str)
            literal_value = golang_common.string_literal(literal.value)
            literal_codes.append(f"{literal_value}: struct{{}}{{}}")

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        errors.append(
            Error(
                constant.parsed.node,
                f"Golang does not support sets of byte arrays, so we can not transpile "
                f"the constant set {constant.name!r}",
            )
        )

    else:
        assert_never(constant.a_type)

    literals_joined = ",\n".join(literal_codes)

    constant_name = golang_naming.constant_name(constant.name)

    writer.write(
        f"""\
var {constant_name} = {set_type} {{
{I}{indent_but_first_line(literals_joined, I)},
}}"""
    )

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_enumeration_literals(
    constant: intermediate.ConstantSetOfEnumerationLiterals,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of enumeration literals."""
    writer = io.StringIO()

    if constant.description is not None:
        comment, comment_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=golang_description.Context(
                package=golang_common.CONSTANTS_PACKAGE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            return None, Error(
                constant.parsed.node,
                f"Failed to generate the documentation comment for {constant.name!r}",
                comment_errors,
            )

        assert comment is not None
        writer.write(comment)
        writer.write("\n")

    enum_name = golang_naming.enum_name(constant.enumeration.name)

    literal_codes = []  # type: List[str]

    for literal in constant.literals:
        literal_name = golang_naming.enum_literal_name(
            constant.enumeration.name, literal.name
        )

        literal_codes.append(f"aastypes.{literal_name}: struct{{}}{{}}")

    constant_name = golang_naming.constant_name(constant.name)

    literals_joined = ",\n".join(literal_codes)

    writer.write(
        f"""\
var {constant_name} = map[aastypes.{enum_name}]struct{{}} {{
{I}{indent_but_first_line(literals_joined, I)},
}}"""
    )

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable, repo_url: Stripped
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the Golang code of the constants based on the symbol table."""
    errors = []  # type: List[Error]

    aastypes_url_literal = golang_common.string_literal(f"{repo_url}/types")

    blocks = [
        Stripped(
            """\
// Package constants provides immutable values of the meta-model.
//
// These are not necessarily constants in the sense of Go language, but
// variables that are expected not to be mutated by the clients.
package constants"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}aastypes {aastypes_url_literal}
)"""
        ),
    ]  # type: List[Stripped]

    for constant in symbol_table.constants:
        block: Optional[Stripped]
        error: Optional[Error]

        if isinstance(constant, intermediate.ConstantPrimitive):
            block, error = _generate_constant_primitive(constant=constant)
        elif isinstance(constant, intermediate.ConstantSetOfPrimitives):
            block, error = _generate_constant_set_of_primitives(constant=constant)
        elif isinstance(constant, intermediate.ConstantSetOfEnumerationLiterals):
            block, error = _generate_constant_set_of_enumeration_literals(
                constant=constant
            )
        else:
            assert_never(constant)

        if error is not None:
            errors.append(error)
            continue

        assert block is not None
        blocks.append(block)

    if len(errors) > 0:
        return None, errors

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# endregion



"""Generate Python code for de-serialization of enumerations."""

import io
from typing import Tuple, Optional, List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Error, Stripped, Identifier
from aas_core_codegen.python import common as python_common, naming as python_naming
from aas_core_codegen.python.common import INDENT as I, INDENT2 as II


def _generate_enum_from_string(
    enumeration: intermediate.Enumeration,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Stripped:
    """Generate the functions for de-serializing enumeration from strings."""
    blocks = []  # type: List[Stripped]

    name = python_naming.enum_name(enumeration.name)

    # region From-string-map

    from_str_map_name = python_naming.constant_name(
        Identifier(f"_{enumeration.name}_from_str")
    )

    from_str_map_writer = io.StringIO()
    from_str_map_writer.write(
        f"""\
{from_str_map_name}: Mapping[str, aas_types.{name}] = {{
"""
    )

    for literal in enumeration.literals:
        literal_name = python_naming.enum_literal_name(literal.name)
        from_str_map_writer.write(
            f"{I}{python_common.string_literal(literal.value)}: "
            f"aas_types.{name}.{literal_name},\n"
        )

    from_str_map_writer.write("}")

    blocks.append(Stripped(from_str_map_writer.getvalue()))

    # endregion

    # region From-string-function

    from_str_name = python_naming.function_name(
        Identifier(f"{enumeration.name}_from_str")
    )

    from_str_writer = io.StringIO()
    from_str_writer.write(
        f"""\
def {from_str_name}(
{II}text: str
) -> Optional[aas_types.{name}]:
{I}\"\"\"
{I}Parse :paramref:`text` as string representation
{I}of :py:class:`{qualified_module_name}.{name}`.

{I}If :paramref:`text` is not a valid string representation of a literal
{I}of :py:class:`{qualified_module_name}.{name}`, return ``None``.

{I}:param text: to be parsed
{I}:return:
{II}the corresponding literal of :py:class:`{qualified_module_name}.{name}`
{II}or ``None``, if :paramref:`text` invalid.
{I}\"\"\"
{I}return {from_str_map_name}.get(text, None)"""
    )

    blocks.append(Stripped(from_str_writer.getvalue()))

    # endregion

    return Stripped("\n\n\n".join(blocks))


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the Python code for the de-serialization of strings.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    blocks = [
        Stripped('"""De-serialize enumerations from string representations."""'),
        python_common.WARNING,
        Stripped(
            f"""\
from typing import (
{I}Mapping,
{I}Optional,
)

import {qualified_module_name}.types as aas_types"""
        ),
    ]

    for enum in symbol_table.enumerations:
        blocks.append(
            _generate_enum_from_string(
                enumeration=enum, qualified_module_name=qualified_module_name
            )
        )

    blocks.append(python_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None



"""Parse the meta model."""

from aas_core_codegen.parse import _types, _translate, _stringify

TypeAnnotation = _types.TypeAnnotation
AtomicTypeAnnotation = _types.AtomicTypeAnnotation
SelfTypeAnnotation = _types.SelfTypeAnnotation
SubscriptedTypeAnnotation = _types.SubscriptedTypeAnnotation
Description = _types.Description
ConstantPrimitive = _types.ConstantPrimitive
SetLiteral = _types.SetLiteral
ConstantSet = _types.ConstantSet
ConstantUnion = _types.ConstantUnion
Property = _types.Property
Default = _types.Default
Argument = _types.Argument
Invariant = _types.Invariant
Contract = _types.Contract
Snapshot = _types.Snapshot
Contracts = _types.Contracts
is_string_expr = _types.is_string_expr
ImplementationSpecificMethod = _types.ImplementationSpecificMethod
UnderstoodMethod = _types.UnderstoodMethod
ConstructorToBeUnderstood = _types.ConstructorToBeUnderstood
Method = _types.Method
MethodUnion = _types.MethodUnion
FunctionUnion = _types.FunctionUnion
OurType = _types.OurType
Serialization = _types.Serialization
Class = _types.Class
AbstractClass = _types.AbstractClass
ClassUnion = _types.ClassUnion
ConcreteClass = _types.ConcreteClass
EnumerationLiteral = _types.EnumerationLiteral
Enumeration = _types.Enumeration
MetaModel = _types.MetaModel
SymbolTable = _types.SymbolTable

PRIMITIVE_TYPES = _types.PRIMITIVE_TYPES
GENERIC_TYPES = _types.GENERIC_TYPES

source_to_atok = _translate.source_to_atok
check_expected_imports = _translate.check_expected_imports
atok_to_symbol_table = _translate.atok_to_symbol_table

dump = _stringify.dump



"""Generate C# code for XML-ization based on the intermediate representation."""

import io
import textwrap
from typing import Tuple, Optional, List

from icontract import ensure, require

from aas_core_codegen import intermediate, naming, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    assert_never,
    indent_but_first_line,
)
from aas_core_codegen.csharp import (
    common as csharp_common,
    naming as csharp_naming,
)
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


def _generate_skip_whitespace_and_comments() -> Stripped:
    """Generate the function to skip whitespace text and XML comments."""
    return Stripped(
        f"""\
internal static void SkipNoneWhitespaceAndComments(
{I}Xml.XmlReader reader)
{{
{I}while (
{II}!reader.EOF
{II}&& (
{III}reader.NodeType == Xml.XmlNodeType.None
{III}|| reader.NodeType == Xml.XmlNodeType.Whitespace
{III}|| reader.NodeType == Xml.XmlNodeType.Comment))
{I}{{
{II}reader.Read();
{I}}}
}}"""
    )


def _generate_read_whole_content_as_base_64() -> Stripped:
    """Generate the function to read the whole of element's content as bytes."""
    return Stripped(
        f"""\
/// <summary>
/// Read the whole content of an element into memory.
/// </summary>
private static byte[] ReadWholeContentAsBase64(
{I}Xml.XmlReader reader)
{{
{I}// The capacity of 1024 bytes is an arbitrary,
{I}// but plausible default capacity.
{I}byte[] buffer = new byte[1024];
{I}using System.IO.MemoryStream stream = (
{II}new System.IO.MemoryStream(1024));
{I}int readBytes;
{I}while ((readBytes = reader.ReadContentAsBase64(buffer, 0, 1024)) > 0)
{I}{{
{II}stream.Write(buffer, 0, readBytes);
{I}}}
{I}return stream.ToArray();
}}"""
    )


def _generate_extract_element_name() -> Stripped:
    """Generate the function to strip the prefix and check the namespace."""
    return Stripped(
        f"""\
/// <summary>
/// Check the namespace and extract the element's name.
/// </summary>
private static string TryElementName(
{I}Xml.XmlReader reader,
{I}out Reporting.Error? error
{I})
{{
{I}// Pre-condition
{I}if (reader.NodeType != Xml.XmlNodeType.Element
{II}&& reader.NodeType != Xml.XmlNodeType.EndElement)
{I}{{
{II}throw new System.InvalidOperationException(
{III}"Expected to be at a start or an end element " +
{III}$"in {{nameof(TryElementName)}}, " +
{III}$"but got: {{reader.NodeType}}");
{I}}}

{I}error = null;
{I}if (reader.NamespaceURI != NS)
{I}{{
{II}error = new Reporting.Error(
{III}$"Expected an element within a namespace {{NS}}, " +
{III}$"but got: {{reader.NamespaceURI}}");
{III}return "";
{I}}}

{I}return reader.LocalName;
}}"""
    )


def _generate_deserialize_primitive_property(
    prop: intermediate.Property, cls: intermediate.ConcreteClass
) -> Stripped:
    """Generate the snippet to deserialize a property ``prop`` of primitive type."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    a_type = intermediate.try_primitive_type(type_anno)
    assert a_type is not None, f"Unexpected type annotation: {prop.type_annotation}"

    deserialization_expr: str
    if a_type is intermediate.PrimitiveType.BOOL:
        deserialization_expr = "reader.ReadContentAsBoolean()"
    elif a_type is intermediate.PrimitiveType.INT:
        deserialization_expr = "reader.ReadContentAsLong()"
    elif a_type is intermediate.PrimitiveType.FLOAT:
        deserialization_expr = "reader.ReadContentAsDouble()"
    elif a_type is intermediate.PrimitiveType.STR:
        deserialization_expr = "reader.ReadContentAsString()"
    elif a_type is intermediate.PrimitiveType.BYTEARRAY:
        deserialization_expr = f"""\
DeserializeImplementation.ReadWholeContentAsBase64(
{I}reader)"""
    else:
        assert_never(a_type)

    target_var = csharp_naming.variable_name(Identifier(f"the_{prop.name}"))

    prop_name = csharp_naming.property_name(prop.name)
    cls_name = csharp_naming.class_name(cls.name)
    xml_prop_name_literal = csharp_common.string_literal(naming.xml_property(prop.name))

    if a_type is intermediate.PrimitiveType.STR:
        empty_handling_body = Stripped(f'{target_var} = "";')
    else:
        empty_handling_body = Stripped(
            f"""\
error = new Reporting.Error(
{I}"The property {prop_name} of an instance of class {cls_name} " +
{I}"can not be de-serialized from a self-closing element " +
{I}"since it needs content");
error.PrependSegment(
{I}new Reporting.NameSegment(
{II}{xml_prop_name_literal}));
return null;"""
        )

    return Stripped(
        f"""\
if (isEmptyProperty)
{{
{I}{indent_but_first_line(empty_handling_body, I)}
}}
else
{{
{I}if (reader.EOF)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected an XML content representing " +
{III}"the property {prop_name} of an instance of class {cls_name}, " +
{III}"but reached the end-of-file");
{II}return null;
{I}}}

{I}try
{I}{{
{II}{target_var} = {indent_but_first_line(deserialization_expr, I)};
{I}}}
{I}catch (System.Exception exception)
{I}{{
{II}if (exception is System.FormatException
{III}|| exception is System.Xml.XmlException)
{II}{{
{III}error = new Reporting.Error(
{IIII}"The property {prop_name} of an instance of class {cls_name} " +
{IIII}$"could not be de-serialized: {{exception.Message}}");
{III}error.PrependSegment(
{IIII}new Reporting.NameSegment(
{IIIII}{xml_prop_name_literal}));
{III}return null;
{II}}}

{II}throw;
{I}}}
}}"""
    )


def _generate_deserialize_enumeration_property(
    prop: intermediate.Property, cls: intermediate.ConcreteClass
) -> Stripped:
    """Generate the snippet to deserialize a property ``prop`` as an enum."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    assert isinstance(type_anno, intermediate.OurTypeAnnotation)

    our_type = type_anno.our_type
    assert isinstance(our_type, intermediate.Enumeration)

    target_var = csharp_naming.variable_name(Identifier(f"the_{prop.name}"))
    text_var = csharp_naming.variable_name(Identifier(f"text_{prop.name}"))

    prop_name = csharp_naming.property_name(prop.name)
    cls_name = csharp_naming.class_name(cls.name)
    enum_name = csharp_naming.enum_name(our_type.name)
    xml_prop_name_literal = csharp_common.string_literal(naming.xml_property(prop.name))

    return Stripped(
        f"""\
if (isEmptyProperty)
{{
{I}error = new Reporting.Error(
{II}"The property {prop_name} of an instance of class {cls_name} " +
{II}"can not be de-serialized from a self-closing element " +
{II}"since it needs content");
{I}error.PrependSegment(
{II}new Reporting.NameSegment(
{III}{xml_prop_name_literal}));
{I}return null;
}}

{I}if (reader.EOF)
{{
{II}error = new Reporting.Error(
{III}"Expected an XML content representing " +
{III}"the property {prop_name} of an instance of class {cls_name}, " +
{III}"but reached the end-of-file");
{II}return null;
}}

string {text_var};
try
{{
{I}{text_var} = reader.ReadContentAsString();
}}
catch (System.FormatException exception)
{{
{I}error = new Reporting.Error(
{II}"The property {prop_name} of an instance of class {cls_name} " +
{II}$"could not be de-serialized as a string: {{exception}}");
{I}error.PrependSegment(
{II}new Reporting.NameSegment(
{III}{xml_prop_name_literal}));
{I}return null;
}}

{target_var} = Stringification.{enum_name}FromString(
{I}{text_var});

if ({target_var} == null)
{{
{I}error = new Reporting.Error(
{II}"The property {prop_name} of an instance of class {cls_name} " +
{II}"could not be de-serialized from an unexpected enumeration literal: " +
{II}{text_var});
{I}error.PrependSegment(
{II}new Reporting.NameSegment(
{III}{xml_prop_name_literal}));
{I}return null;
}}"""
    )


def _generate_deserialize_interface_property(
    prop: intermediate.Property,
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """Generate the snippet to deserialize a property ``prop`` as an interface."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    assert isinstance(type_anno, intermediate.OurTypeAnnotation)

    our_type = type_anno.our_type
    assert isinstance(
        our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
    )
    assert our_type.interface is not None

    prop_name = csharp_naming.property_name(prop.name)
    cls_name = csharp_naming.class_name(cls.name)

    interface_name = csharp_naming.interface_name(our_type.interface.name)

    target_var = csharp_naming.variable_name(Identifier(f"the_{prop.name}"))
    xml_prop_name_literal = csharp_common.string_literal(naming.xml_property(prop.name))

    return Stripped(
        f"""\
if (isEmptyProperty)
{{
{I}error = new Reporting.Error(
{II}$"Expected an XML element within the element {{elementName}} representing " +
{II}"the property {prop_name} of an instance of class {cls_name}, " +
{II}"but encountered a self-closing element {{elementName}}");
{I}return null;
}}

// We need to skip the whitespace here in order to be able to look ahead
// the discriminator element shortly.
SkipNoneWhitespaceAndComments(reader);

if (reader.EOF)
{{
{I}error = new Reporting.Error(
{II}$"Expected an XML element within the element {{elementName}} representing " +
{II}"the property {prop_name} of an instance of class {cls_name}, " +
{II}"but reached the end-of-file");
{I}return null;
}}

// Try to look ahead the discriminator name;
// we need this name only for the error reporting below.
// {interface_name}FromElement will perform more sophisticated
// checks.
string? discriminatorElementName = null;
if (reader.NodeType == Xml.XmlNodeType.Element)
{{
{I}discriminatorElementName = reader.LocalName;
}}

{target_var} = {interface_name}FromElement(
{I}reader, out error);

if (error != null)
{{
{I}if (discriminatorElementName != null)
{I}{{
{II}error.PrependSegment(
{III}new Reporting.NameSegment(
{IIII}discriminatorElementName));
{I}}}

{I}error.PrependSegment(
{II}new Reporting.NameSegment(
{III}{xml_prop_name_literal}));
{I}return null;
}}"""
    )


def _generate_deserialize_cls_property(prop: intermediate.Property) -> Stripped:
    """Generate the snippet to deserialize a property ``prop`` as a concrete class."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    assert isinstance(type_anno, intermediate.OurTypeAnnotation)

    our_type = type_anno.our_type
    assert isinstance(our_type, intermediate.ConcreteClass)

    target_cls_name = csharp_naming.class_name(our_type.name)

    target_var = csharp_naming.variable_name(Identifier(f"the_{prop.name}"))
    xml_prop_name_literal = csharp_common.string_literal(naming.xml_property(prop.name))

    return Stripped(
        f"""\
{target_var} = {target_cls_name}FromSequence(
{I}reader, isEmptyProperty, out error);

if (error != null)
{{
{I}error.PrependSegment(
{II}new Reporting.NameSegment(
{III}{xml_prop_name_literal}));
{I}return null;
}}"""
    )


def _generate_deserialize_list_property(prop: intermediate.Property) -> Stripped:
    """Generate the code to de-serialize a property ``prop`` as a list."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    # fmt: off
    assert (
        isinstance(type_anno, intermediate.ListTypeAnnotation)
        and isinstance(type_anno.items, intermediate.OurTypeAnnotation)
        and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass)
        )
    ), "See intermediate._translate._verify_only_simple_type_patterns"
    # fmt: on

    target_var = csharp_naming.variable_name(Identifier(f"the_{prop.name}"))
    index_var = csharp_naming.variable_name(Identifier(f"index_{prop.name}"))

    item_our_type = type_anno.items.our_type
    if (
        isinstance(item_our_type, intermediate.AbstractClass)
        or len(item_our_type.concrete_descendants) > 0
    ):
        deserialize_method = (
            f"{csharp_naming.interface_name(type_anno.items.our_type.name)}FromElement"
        )
    else:
        deserialize_method = (
            f"{csharp_naming.class_name(type_anno.items.our_type.name)}FromElement"
        )

    item_type = csharp_common.generate_type(type_anno.items)

    xml_prop_name = naming.xml_property(prop.name)
    xml_prop_name_literal = csharp_common.string_literal(xml_prop_name)

    body_for_non_empty_property = Stripped(
        f"""\
SkipNoneWhitespaceAndComments(reader);

int {index_var} = 0;
while (reader.NodeType == Xml.XmlNodeType.Element)
{{
{I}{item_type}? item = {deserialize_method}(
{II}reader, out error);

{I}if (error != null)
{I}{{
{II}error.PrependSegment(
{III}new Reporting.IndexSegment(
{IIII}{index_var}));
error.PrependSegment(
{I}new Reporting.NameSegment(
{II}{xml_prop_name_literal}));
{II}return null;
{I}}}

{I}{target_var}.Add(
{II}item
{III}?? throw new System.InvalidOperationException(
{IIII}"Unexpected item null when error null"));

{I}{index_var}++;
{I}SkipNoneWhitespaceAndComments(reader);
}}"""
    )

    return Stripped(
        f"""\
{target_var} = new List<{item_type}>();

if (!isEmptyProperty)
{{
{I}{indent_but_first_line(body_for_non_empty_property, I)}
}}"""
    )


@require(lambda prop, cls: id(prop) in cls.property_id_set)
def _generate_deserialize_property(
    prop: intermediate.Property, cls: intermediate.ConcreteClass
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the snippet to deserialize the property ``prop`` from the content."""
    blocks = []  # type: List[Stripped]

    type_anno = intermediate.beneath_optional(prop.type_annotation)

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        blocks.append(_generate_deserialize_primitive_property(prop=prop, cls=cls))
    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        our_type = type_anno.our_type
        if isinstance(our_type, intermediate.Enumeration):
            blocks.append(
                _generate_deserialize_enumeration_property(prop=prop, cls=cls)
            )
        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2022-04-13):
            # The constrained primitives are only verified, but not represented as
            # separate classes in the XSD.
            blocks.append(_generate_deserialize_primitive_property(prop=prop, cls=cls))
        elif isinstance(
            our_type, (intermediate.ConcreteClass, intermediate.AbstractClass)
        ):
            if (
                isinstance(our_type, intermediate.AbstractClass)
                or len(our_type.concrete_descendants) > 0
            ):
                blocks.append(
                    _generate_deserialize_interface_property(prop=prop, cls=cls)
                )
            else:
                blocks.append(_generate_deserialize_cls_property(prop=prop))
        else:
            assert_never(our_type)

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        blocks.append(_generate_deserialize_list_property(prop=prop))

    else:
        assert_never(type_anno)

    return Stripped("\n\n".join(blocks)), None


def _generate_deserialize_impl_cls_from_sequence(
    cls: intermediate.ConcreteClass,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the function to de-serialize the ``cls`` from an XML sequence."""
    name = csharp_naming.class_name(identifier=cls.name)

    description = Stripped(
        f"""\
/// <summary>
/// Deserialize an instance of class {name} from a sequence of XML elements.
/// </summary>
/// <remarks>
/// If <paramref name="isEmptySequence" /> is set, we should try to deserialize
/// the instance from an empty sequence. That is, the parent element
/// was a self-closing element.
/// </remarks>"""
    )

    # NOTE (mristin, 2022-06-21):
    # Hard-wire for the case when no sequence is read
    if len(cls.constructor.arguments) == 0:
        return (
            Stripped(
                f"""\
{description}
internal static Aas.{name} {name}FromSequence(
{I}Xml.XmlReader reader,
{I}bool isEmptySequence,
{I}out Reporting.Error? error)
{{
{I}error = null;
{I}return new Aas.{name}();
}}  // internal static Aas.{name}? {name}FromSequence"""
            ),
            None,
        )

    errors = []  # type: List[Error]

    blocks = [
        Stripped("error = null;"),
    ]  # type: List[Stripped]

    assert len(cls.constructor.arguments) > 0, "Otherwise expected hard-wiring above"
    init_target_var_stmts = []  # type: List[Stripped]
    for prop in cls.properties:
        target_type = csharp_common.generate_type(prop.type_annotation)
        target_var = csharp_naming.variable_name(Identifier(f"the_{prop.name}"))

        # NOTE (mristin, 2022-04-13):
        # This is a poor man's trick to make all temporary variables optional.
        # The required constructor arguments / properties will be checked just
        # before the constructor as we can not predict in advance which properties
        # were actually provided without any lookahead in XML reading.
        if not target_type.endswith("?"):
            target_type = Stripped(f"{target_type}?")

        init_target_var_stmts.append(Stripped(f"{target_type} {target_var} = null;"))
    blocks.append(Stripped("\n".join(init_target_var_stmts)))

    # noinspection PyListCreation
    blocks_for_non_empty = []  # type: List[Stripped]

    blocks_for_non_empty.append(
        Stripped(
            f"""\
SkipNoneWhitespaceAndComments(reader);
if (reader.EOF)
{{
{I}error = new Reporting.Error(
{II}"Expected an XML element representing " +
{II}"a property of an instance of class {name}, " +
{II}"but reached the end-of-file");
{I}return null;
}}"""
        )
    )

    case_blocks = []  # type: List[Stripped]
    for prop in cls.properties:
        case_body, error = _generate_deserialize_property(prop=prop, cls=cls)
        if error is not None:
            errors.append(error)
            continue

        assert case_body is not None

        xml_prop_name = naming.xml_property(prop.name)
        xml_prop_name_literal = csharp_common.string_literal(xml_prop_name)
        case_blocks.append(
            Stripped(
                f"""\
case {xml_prop_name_literal}:
{{
{I}{indent_but_first_line(case_body, I)}
{I}break;
}}"""
            )
        )

    if len(errors) > 0:
        return None, errors

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}error = new Reporting.Error(
{II}"We expected properties of the class {name}, " +
{II}"but got an unexpected element " +
{II}$"with the name {{elementName}}");
{I}return null;"""
        )
    )

    switch_body = "\n".join(case_blocks)

    blocks_for_non_empty.append(
        Stripped(
            f"""\
while (true)
{{
{I}SkipNoneWhitespaceAndComments(reader);

{I}if (reader.NodeType == Xml.XmlNodeType.EndElement || reader.EOF)
{I}{{
{II}break;
{I}}}

{I}if (reader.NodeType != Xml.XmlNodeType.Element)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected an XML start element representing " +
{III}"a property of an instance of class {name}, " +
{III}$"but got the node of type {{reader.NodeType}} " +
{III}$"with the value {{reader.Value}}");
{II}return null;
{I}}}

{I}string elementName = TryElementName(
{II}reader, out error);
{I}if (error != null)
{I}{{
{II}return null;
{I}}}

{I}bool isEmptyProperty = reader.IsEmptyElement;

{I}// Skip the expected element
{I}reader.Read();

{I}switch (elementName)
{I}{{
{II}{indent_but_first_line(switch_body, II)}
{I}}}

{I}SkipNoneWhitespaceAndComments(reader);

{I}if (!isEmptyProperty)
{I}{{
{II}// Read the end element

{II}if (reader.EOF)
{II}{{
{III}error = new Reporting.Error(
{IIII}"Expected an XML end element to conclude a property of class {name} " +
{IIII}$"with the element name {{elementName}}, " +
{IIII}"but got the end-of-file.");
{III}return null;
{II}}}
{II}if (reader.NodeType != Xml.XmlNodeType.EndElement)
{II}{{
{III}error = new Reporting.Error(
{IIII}"Expected an XML end element to conclude a property of class {name} " +
{IIII}$"with the element name {{elementName}}, " +
{IIII}$"but got the node of type {{reader.NodeType}} " +
{IIII}$"with the value {{reader.Value}}");
{III}return null;
{II}}}

{II}string endElementName = TryElementName(
{III}reader, out error);
{II}if (error != null)
{II}{{
{III}return null;
{II}}}

{II}if (endElementName != elementName)
{II}{{
{III}error = new Reporting.Error(
{IIII}"Expected an XML end element to conclude a property of class {name} " +
{IIII}$"with the element name {{elementName}}, " +
{IIII}$"but got the end element with the name {{reader.Name}}");
{III}return null;
{II}}}
{II}// Skip the expected end element
{II}reader.Read();
{I}}}
}}"""
        )
    )

    body_for_non_empty_sequence = "\n".join(blocks_for_non_empty)
    blocks.append(
        Stripped(
            f"""\
if (!isEmptySequence)
{{
{I}{indent_but_first_line(body_for_non_empty_sequence, I)}
}}"""
        )
    )

    # region Check that the mandatory properties have been set

    for prop in cls.properties:
        prop_csharp = csharp_naming.property_name(prop.name)
        target_var = csharp_naming.variable_name(Identifier(f"the_{prop.name}"))

        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            blocks.append(
                Stripped(
                    f"""\
if ({target_var} == null)
{{
{I}error = new Reporting.Error(
{II}"The required property {prop_csharp} has not been given " +
{II}"in the XML representation of an instance of class {name}");
{I}return null;
}}"""
                )
            )

    # endregion

    # region Pass in properties as arguments to the constructor

    property_names = [prop.name for prop in cls.properties]
    constructor_argument_names = [arg.name for arg in cls.constructor.arguments]

    # fmt: off
    assert (
            set(prop.name for prop in cls.properties)
            == set(arg.name for arg in cls.constructor.arguments)
    ), (
        f"Expected the properties to coincide with constructor arguments, "
        f"but they do not for {cls.name!r}:"
        f"{property_names=}, {constructor_argument_names=}"
    )
    # fmt: on

    init_writer = io.StringIO()
    init_writer.write(f"return new Aas.{name}(\n")

    for i, arg in enumerate(cls.constructor.arguments):
        prop = cls.properties_by_name[arg.name]

        # NOTE (mristin, 2022-04-13):
        # The argument to the constructor may be optional while the property might
        # be required, since we can set the default value in the body of the
        # constructor. However, we can not have an optional property and a required
        # constructor argument as we then would not know how to create the instance.

        if not (
            intermediate.type_annotations_equal(
                arg.type_annotation, prop.type_annotation
            )
            or intermediate.type_annotations_equal(
                intermediate.beneath_optional(arg.type_annotation),
                prop.type_annotation,
            )
        ):
            errors.append(
                Error(
                    arg.parsed.node,
                    f"Expected type annotation for property {prop.name!r} "
                    f"and constructor argument {arg.name!r} "
                    f"of the class {cls.name!r} to have matching types, "
                    f"but they do not: "
                    f"property type is {prop.type_annotation} "
                    f"and argument type is {arg.type_annotation}. "
                    f"Hence we do not know how to generate the call "
                    f"to the constructor in the JSON de-serialization.",
                )
            )
            continue

        arg_var = csharp_naming.variable_name(Identifier(f"the_{arg.name}"))

        init_writer.write(f"{I}{arg_var}")
        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            init_writer.write("\n")

            # Dedention could not work here due to prefix indention at the very
            # beginning.
            init_writer.write(
                f"""\
{II} ?? throw new System.InvalidOperationException(
{III}"Unexpected null, had to be handled before")"""
            )

        if i < len(cls.constructor.arguments) - 1:
            init_writer.write(",\n")
        else:
            init_writer.write(");")

    if len(errors) > 0:
        return None, errors

    # endregion

    blocks.append(Stripped(init_writer.getvalue()))

    writer = io.StringIO()
    writer.write(
        f"""\
{description}
internal static Aas.{name}? {name}FromSequence(
{I}Xml.XmlReader reader,
{I}bool isEmptySequence,
{I}out Reporting.Error? error)
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write(f"\n}}  // internal static Aas.{name}? {name}FromSequence")

    return Stripped(writer.getvalue()), None


def _generate_deserialize_impl_concrete_cls_from_element(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """Generate the function to de-serialize a concrete ``cls`` from an XML element."""
    name = csharp_naming.class_name(cls.name)
    xml_name = naming.xml_class_name(cls.name)
    xml_name_literal = csharp_common.string_literal(xml_name)

    # NOTE (mristin, 2022-06-21):
    # We need to propagate nullability. Otherwise, InspectCode complains.
    result_nullability = "?" if len(cls.constructor.arguments) > 0 else ""

    body = Stripped(
        f"""\
error = null;

SkipNoneWhitespaceAndComments(reader);

if (reader.EOF)
{{
{I}error = new Reporting.Error(
{II}"Expected an XML element representing an instance of class {name}, " +
{II}"but reached the end-of-file");
{I}return null;
}}

if (reader.NodeType != Xml.XmlNodeType.Element)
{{
{I}error = new Reporting.Error(
{II}"Expected an XML element representing an instance of class {name}, " +
{II}$"but got a node of type {{reader.NodeType}} " +
{II}$"with value {{reader.Value}}");
{I}return null;
}}

string elementName = TryElementName(
    reader, out error);
if (error != null)
{{
{I}return null;
}}

if (elementName != {xml_name_literal})
{{
{I}error = new Reporting.Error(
{II}"Expected an element representing an instance of class {name} " +
{II}$"with element name {xml_name}, but got: {{elementName}}");
{I}return null;
}}

bool isEmptyElement = reader.IsEmptyElement;

// Skip the element node and go to the content
reader.Read();

Aas.{name}{result_nullability} result = (
{I}{name}FromSequence(
{II}reader, isEmptyElement, out error));
if (error != null)
{{
{I}return null;
}}

SkipNoneWhitespaceAndComments(reader);

if (!isEmptyElement)
{{
{I}if (reader.EOF)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected an XML end element concluding an instance of class {name}, " +
{III}"but reached the end-of-file");
{II}return null;
{I}}}

{I}if (reader.NodeType != Xml.XmlNodeType.EndElement)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected an XML end element concluding an instance of class {name}, " +
{III}$"but got a node of type {{reader.NodeType}} " +
{III}$"with value {{reader.Value}}");
{II}return null;
{I}}}

{I}string endElementName = TryElementName(
{II}reader, out error);
{I}if (error != null)
{I}{{
{II}return null;
{I}}}

{I}if (endElementName != elementName)
{I}{{
{II}error = new Reporting.Error(
{III}$"Expected an XML end element with an name {{elementName}}, " +
{III}$"but got: {{endElementName}}");
{II}return null;
{I}}}

{I}// Skip the end element
{I}reader.Read();
}}

return result;"""
    )

    return Stripped(
        f"""\
/// <summary>
/// Deserialize an instance of class {name} from an XML element.
/// </summary>
internal static Aas.{name}? {name}FromElement(
{I}Xml.XmlReader reader,
{I}out Reporting.Error? error)
{{
{I}{indent_but_first_line(body, I)}
}}  // internal static Aas.{name}? {name}FromElement"""
    )


def _generate_deserialize_impl_interface_from_element(
    interface: intermediate.Interface,
) -> Stripped:
    """Generate the function to de-serialize an ``interface`` from an XML element."""
    name = csharp_naming.interface_name(interface.name)

    blocks = [
        Stripped(
            f"""\
error = null;

SkipNoneWhitespaceAndComments(reader);

if (reader.EOF)
{{
{I}error = new Reporting.Error(
{II}"Expected an XML element, but reached end-of-file");
{I}return null;
}}

if (reader.NodeType != Xml.XmlNodeType.Element)
{{
{I}error = new Reporting.Error(
{II}"Expected an XML element, " +
{II}$"but got a node of type {{reader.NodeType}} " +
{II}$"with value {{reader.Value}}");
{I}return null;
}}"""
        )
    ]  # type: List[Stripped]

    case_stmts = []  # type: List[Stripped]
    for implementer in interface.implementers:
        implementer_xml_name_literal = csharp_common.string_literal(
            naming.xml_class_name(implementer.name)
        )

        implementer_name = csharp_naming.class_name(implementer.name)

        case_stmts.append(
            Stripped(
                f"""\
case {implementer_xml_name_literal}:
{I}return {implementer_name}FromElement(
{II}reader, out error);"""
            )
        )

    case_stmts.append(
        Stripped(
            f"""\
default:
{I}error = new Reporting.Error(
{II}$"Unexpected element with the name {{elementName}}");
{I}return null;"""
        )
    )

    switch_writer = io.StringIO()
    switch_writer.write(
        f"""\
string elementName = TryElementName(
{I}reader, out error);
if (error != null)
{{
{I}return null;
}}

switch (elementName)
{{
"""
    )
    for i, case_stmt in enumerate(case_stmts):
        if i > 0:
            switch_writer.write("\n")
        switch_writer.write(textwrap.indent(case_stmt, I))

    switch_writer.write("\n}")

    blocks.append(Stripped(switch_writer.getvalue()))

    writer = io.StringIO()
    writer.write(
        f"""\
/// <summary>
/// Deserialize an instance of {name} from an XML element.
/// </summary>
[CodeAnalysis.SuppressMessage("ReSharper", "InconsistentNaming")]
internal static Aas.{name}? {name}FromElement(
{I}Xml.XmlReader reader,
{I}out Reporting.Error? error)
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write(f"\n}}  // internal static Aas.{name}? {name}FromElement")

    return Stripped(writer.getvalue())


def _generate_deserialize_impl(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the implementation for deserialization functions."""
    blocks = [
        _generate_skip_whitespace_and_comments(),
        _generate_read_whole_content_as_base_64(),
        _generate_extract_element_name(),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    # NOTE (mristin, 2022-04-13):
    # Enumerations are going to be directly deserialized using
    # ``Stringification``.

    # NOTE (mristin, 2022-04-13):
    # Constrained primitives are only verified, but do not represent a C# type.

    for cls in symbol_table.classes:
        if cls.is_implementation_specific:
            implementation_keys = [
                specific_implementations.ImplementationKey(
                    f"Xmlization/DeserializeImplementation/"
                    f"{cls.name}_from_element.cs"
                ),
                specific_implementations.ImplementationKey(
                    f"Xmlization/DeserializeImplementation/"
                    f"{cls.name}_from_sequence.cs"
                ),
            ]

            for implementation_key in implementation_keys:
                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            cls.parsed.node,
                            f"The xmlization snippet is missing "
                            f"for the implementation-specific "
                            f"class {cls.name}: {implementation_key}",
                        )
                    )
                    continue
                else:
                    blocks.append(spec_impls[implementation_key])
        else:
            if isinstance(cls, intermediate.ConcreteClass):
                (
                    block,
                    generation_errors,
                ) = _generate_deserialize_impl_cls_from_sequence(cls=cls)
                if generation_errors is not None:
                    errors.append(
                        Error(
                            cls.parsed.node,
                            f"Failed to generate the XML deserialization code "
                            f"for the class {cls.name}",
                            generation_errors,
                        )
                    )
                else:
                    assert block is not None
                    blocks.append(block)

            if cls.interface is not None:
                blocks.append(
                    _generate_deserialize_impl_interface_from_element(
                        interface=cls.interface
                    )
                )

            if isinstance(cls, intermediate.ConcreteClass):
                blocks.append(
                    _generate_deserialize_impl_concrete_cls_from_element(cls=cls)
                )

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()

    writer.write(
        """\
/// <summary>
/// Implement the deserialization of meta-model classes from XML.
/// </summary>
/// <remarks>
/// The implementation propagates an <see cref="Reporting.Error" /> instead of
/// relying on exceptions. Under the assumption that incorrect data is much less
/// frequent than correct data, this makes the deserialization more
/// efficient.
///
/// However, we do not want to force the client to deal with
/// the <see cref="Reporting.Error" /> class as this is not intuitive.
/// Therefore we distinguish the implementation, realized in
/// <see cref="DeserializeImplementation" />, and the facade given in
/// <see cref="Deserialize" /> class.
/// </remarks>
internal static class DeserializeImplementation
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // internal static class DeserializeImplementation")

    return Stripped(writer.getvalue()), None


def _generate_deserialize_from(name: Identifier) -> Stripped:
    """Generate the facade method for deserialization of the class or interface."""
    writer = io.StringIO()

    writer.write(
        f"""\
/// <summary>
/// Deserialize an instance of {name} from <paramref name="reader" />.
/// </summary>
/// <param name="reader">Initialized XML reader with cursor set to the element</param>
/// <exception cref="Xmlization.Exception">
/// Thrown when the element is not a valid XML
/// representation of {name}.
/// </exception>
"""
    )

    if name.startswith("I"):
        writer.write(
            """\
[CodeAnalysis.SuppressMessage("ReSharper", "InconsistentNaming")]"""
        )

    writer.write(
        f"""\
public static Aas.{name} {name}From(
{I}Xml.XmlReader reader)
{{
{I}DeserializeImplementation.SkipNoneWhitespaceAndComments(reader);

{I}if (!reader.EOF && reader.NodeType == Xml.XmlNodeType.XmlDeclaration)
{I}{{
{II}throw new Xmlization.Exception(
{III}"",
{III}"Unexpected XML declaration when reading an instance " +
{III}"of class {name}, as we expect the reader " +
{III}"to be set at content with MoveToContent");
{I}}}

{I}Aas.{name}? result = (
{II}DeserializeImplementation.{name}FromElement(
{III}reader,
{III}out Reporting.Error? error));
{I}if (error != null)
{I}{{
{II}throw new Xmlization.Exception(
{III}Reporting.GenerateRelativeXPath(error.PathSegments),
{III}error.Cause);
{I}}}
{I}return result
{II}?? throw new System.InvalidOperationException(
{III}"Unexpected output null when error is null");
}}"""
    )

    return Stripped(writer.getvalue())


def _generate_deserialize(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the public class ``Deserialize``."""
    blocks = []  # type: List[Stripped]

    # NOTE (mristin, 2022-04-13):
    # We use stringification for de-serialization of enumerations.

    # NOTE (mristin, 2022-04-13):
    # Constrained primitives are not handled as separate classes, but as
    # primitives, and only verified in the verification.

    for cls in symbol_table.classes:
        if cls.interface is not None:
            blocks.append(
                _generate_deserialize_from(
                    name=csharp_naming.interface_name(cls.interface.name)
                )
            )

        if isinstance(cls, intermediate.ConcreteClass):
            blocks.append(
                _generate_deserialize_from(name=csharp_naming.class_name(cls.name))
            )

    writer = io.StringIO()
    writer.write(
        """\
/// <summary>
/// Deserialize instances of meta-model classes from XML.
/// </summary>
"""
    )

    first_cls = symbol_table.classes[0] if len(symbol_table.classes) > 0 else None

    if first_cls is not None:
        cls_name: str
        if isinstance(first_cls, intermediate.AbstractClass):
            cls_name = csharp_naming.interface_name(first_cls.name)
        elif isinstance(first_cls, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(first_cls.name)
        else:
            assert_never(first_cls)

        an_instance_variable = csharp_naming.variable_name(Identifier("an_instance"))

        writer.write(
            f"""\
/// <example>
/// Here is an example how to parse an instance of class {cls_name}:
/// <code>
/// var reader = new System.Xml.XmlReader(/* some arguments */);
/// Aas.{cls_name} {an_instance_variable} = Deserialize.{cls_name}From(
/// {I}reader);
/// </code>
/// </example>
///
/// <example>
/// If the elements live in a namespace, you have to supply it. For example:
/// <code>
/// var reader = new System.Xml.XmlReader(/* some arguments */);
/// Aas.{cls_name} {an_instance_variable} = Deserialize.{cls_name}From(
/// {I}reader,
/// {I}"http://www.example.com/5/12");
/// </code>
/// </example>
"""
        )

    writer.write(
        """\
public static class Deserialize
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public static class Deserialize")

    return Stripped(writer.getvalue())


def _generate_serialize_primitive_property_as_content(
    prop: intermediate.Property,
) -> Stripped:
    """Generate the serialization of the primitive-type ``prop`` as XML content."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    a_type = intermediate.try_primitive_type(type_anno)
    assert (
        a_type is not None
    ), f"Unexpected non-primitive type of the property {prop.name!r}: {type_anno}"

    prop_name = csharp_naming.property_name(prop.name)
    xml_prop_name_literal = csharp_common.string_literal(naming.xml_property(prop.name))

    write_value_block: Stripped

    if (
        a_type is intermediate.PrimitiveType.BOOL
        or a_type is intermediate.PrimitiveType.INT
        or a_type is intermediate.PrimitiveType.FLOAT
        or a_type is intermediate.PrimitiveType.STR
    ):
        write_value_block = Stripped(
            f"""\
writer.WriteValue(
{I}that.{prop_name});"""
        )
    elif a_type is intermediate.PrimitiveType.BYTEARRAY:
        write_value_block = Stripped(
            f"""\
writer.WriteBase64(
{I}that.{prop_name},
{I}0,
{I}that.{prop_name}.Length);"""
        )
    else:
        assert_never(a_type)

    # NOTE (mristin, 2022-06-21):
    # Wrap the write_value_block with property even if we discard it below
    write_value_block = Stripped(
        f"""\
writer.WriteStartElement(
{I}{xml_prop_name_literal},
{I}NS);

{write_value_block}

writer.WriteEndElement();"""
    )

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        if a_type in (
            intermediate.PrimitiveType.BOOL,
            intermediate.PrimitiveType.INT,
            intermediate.PrimitiveType.FLOAT,
        ):
            write_value_block = Stripped(
                f"""\
if (that.{prop_name}.HasValue)
{{
{I}writer.WriteStartElement(
{II}{xml_prop_name_literal},
{II}NS);

{I}writer.WriteValue(
{II}that.{prop_name}.Value);

{I}writer.WriteEndElement();
}}"""
            )
        else:
            write_value_block = Stripped(
                f"""\
if (that.{prop_name} != null)
{{
{I}{indent_but_first_line(write_value_block, I)}
}}"""
            )

    return write_value_block


def _generate_serialize_enumeration_property_as_content(
    prop: intermediate.Property,
) -> Stripped:
    """Generate the serialization of an enumeration ``prop`` as XML content."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)
    assert isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
        type_anno.our_type, intermediate.Enumeration
    ), "See intermediate._translate._verify_only_simple_type_patterns"

    enumeration = type_anno.our_type

    prop_name = csharp_naming.property_name(prop.name)
    xml_prop_name_literal = csharp_common.string_literal(naming.xml_property(prop.name))

    enum_name = csharp_naming.enum_name(enumeration.name)

    text_var = csharp_naming.variable_name(Identifier(f"text_{prop.name}"))
    write_value_block = Stripped(
        f"""\
writer.WriteStartElement(
{I}{xml_prop_name_literal},
{I}NS);

string? {text_var} = Stringification.ToString(
{I}that.{prop_name});
writer.WriteValue(
{I}{text_var}
{II}?? throw new System.ArgumentException(
{III}"Invalid literal for the enumeration {enum_name}: " +
{III}that.{prop_name}.ToString()));

writer.WriteEndElement();"""
    )

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        write_value_block = Stripped(
            f"""\
if (that.{prop_name} != null)
{{
{I}{indent_but_first_line(write_value_block, I)}
}}"""
        )

    return write_value_block


def _generate_serialize_interface_property_as_content(
    prop: intermediate.Property,
) -> Stripped:
    """Generate the serialization of an interface as XML content."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    # fmt: off
    assert (
        isinstance(type_anno, intermediate.OurTypeAnnotation)
        and (
            isinstance(type_anno.our_type, intermediate.AbstractClass)
            or (
                isinstance(type_anno.our_type, intermediate.ConcreteClass)
                and len(type_anno.our_type.concrete_descendants) > 0
            )
        )
    ), "See intermediate._translate._verify_only_simple_type_patterns"
    # fmt: on

    prop_name = csharp_naming.property_name(prop.name)
    xml_prop_name_literal = csharp_common.string_literal(naming.xml_property(prop.name))

    result = Stripped(
        f"""\
writer.WriteStartElement(
{I}{xml_prop_name_literal},
{I}NS);

this.Visit(
{I}that.{prop_name},
{I}writer);

writer.WriteEndElement();"""
    )

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        result = Stripped(
            f"""\
if (that.{prop_name} != null)
{{
{I}{indent_but_first_line(result, I)}
}}"""
        )

    return result


def _generate_serialize_concrete_class_property_as_sequence(
    prop: intermediate.Property,
) -> Stripped:
    """Generate the serialization of the class ``prop`` as a sequence of properties."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)
    assert isinstance(type_anno, intermediate.OurTypeAnnotation)
    assert isinstance(type_anno.our_type, intermediate.ConcreteClass)

    cls_to_sequence = csharp_naming.method_name(
        Identifier(f"{type_anno.our_type.name}_to_sequence")
    )

    prop_name = csharp_naming.property_name(prop.name)
    xml_prop_name_literal = csharp_common.string_literal(naming.xml_property(prop.name))

    result = Stripped(
        f"""\
writer.WriteStartElement(
{I}{xml_prop_name_literal},
{I}NS);

this.{cls_to_sequence}(
{I}that.{prop_name},
{I}writer);

writer.WriteEndElement();"""
    )

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        result = Stripped(
            f"""\
if (that.{prop_name} != null)
{{
{I}{indent_but_first_line(result, I)}
}}"""
        )

    return result


def _generate_serialize_list_property_as_content(
    prop: intermediate.Property,
) -> Stripped:
    """Generate the serialization of a list ``prop`` as a sequence of elements."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    # fmt: off
    assert (
        isinstance(type_anno, intermediate.ListTypeAnnotation)
        and isinstance(type_anno.items, intermediate.OurTypeAnnotation)
        and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass)
        )
    ), "See intermediate._translate._verify_only_simple_type_patterns"
    # fmt: on

    prop_name = csharp_naming.property_name(prop.name)
    xml_prop_name_literal = csharp_common.string_literal(naming.xml_property(prop.name))

    result = Stripped(
        f"""\
writer.WriteStartElement(
{I}{xml_prop_name_literal},
{I}NS);

foreach (var item in that.{prop_name})
{{
{I}this.Visit(
{II}item,
{II}writer);
}}

writer.WriteEndElement();"""
    )

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        result = Stripped(
            f"""\
if (that.{prop_name} != null)
{{
{I}{indent_but_first_line(result, I)}
}}"""
        )

    return result


def _generate_serialize_property_as_content(prop: intermediate.Property) -> Stripped:
    """Generate the code to serialize the ``prop`` as content of an XML element."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    body = None  # type: Optional[Stripped]

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        body = _generate_serialize_primitive_property_as_content(prop=prop)
    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        our_type = type_anno.our_type

        if isinstance(our_type, intermediate.Enumeration):
            body = _generate_serialize_enumeration_property_as_content(prop=prop)

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            body = _generate_serialize_primitive_property_as_content(prop=prop)

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if (
                isinstance(our_type, intermediate.AbstractClass)
                or len(our_type.concrete_descendants) > 0
            ):
                body = _generate_serialize_interface_property_as_content(prop=prop)
            else:
                body = _generate_serialize_concrete_class_property_as_sequence(
                    prop=prop
                )

        else:
            assert_never(our_type)

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        body = _generate_serialize_list_property_as_content(prop=prop)

    else:
        assert_never(type_anno)

    return body


def _generate_class_to_sequence(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the method to write ``cls`` as a sequence of properties as XML."""
    blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        body = _generate_serialize_property_as_content(prop=prop)
        blocks.append(body)

    interface_name = csharp_naming.interface_name(cls.name)
    method_name = csharp_naming.method_name(Identifier(f"{cls.name}_to_sequence"))

    writer = io.StringIO()

    if len(cls.properties) == 0:
        blocks.append(Stripped("// Intentionally empty."))

        writer.write(
            '[CodeAnalysis.SuppressMessage("ReSharper", "UnusedParameter.Local")]\n'
        )

    writer.write(
        f"""\
private void {method_name}(
{I}Aas.{interface_name} that,
{I}Xml.XmlWriter writer)
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write(f"\n}}  // private void {method_name}")

    return Stripped(writer.getvalue())


def _generate_visit_for_class(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the method to write the ``cls`` as an XML element."""
    interface_name = csharp_naming.interface_name(cls.name)
    visit_name = csharp_naming.method_name(Identifier(f"visit_{cls.name}"))

    cls_to_sequence_name = csharp_naming.method_name(
        Identifier(f"{cls.name}_to_sequence")
    )

    xml_cls_name_literal = csharp_common.string_literal(naming.xml_class_name(cls.name))

    return Stripped(
        f"""\
public override void {visit_name}(
{I}Aas.{interface_name} that,
{I}Xml.XmlWriter writer)
{{
{I}writer.WriteStartElement(
{II}{xml_cls_name_literal},
{II}NS);
{I}this.{cls_to_sequence_name}(
{II}that,
{II}writer);
{I}writer.WriteEndElement();
}}"""
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_visitor(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate a visitor which serializes instances of the meta-model to XML."""
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    # The abstract classes are directly dispatched by the transformer,
    # so we do not need to handle them separately.

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_keys = [
                specific_implementations.ImplementationKey(
                    f"Xmlization/VisitorWithWriter/visit_{cls.name}.cs"
                ),
                specific_implementations.ImplementationKey(
                    f"Xmlization/VisitorWithWriter/{cls.name}_to_sequence.cs"
                ),
            ]

            for implementation_key in implementation_keys:
                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            cls.parsed.node,
                            f"The xmlization snippet is missing "
                            f"for the implementation-specific "
                            f"class {cls.name}: {implementation_key}",
                        )
                    )
                    continue

                blocks.append(spec_impls[implementation_key])
        else:
            blocks.append(_generate_class_to_sequence(cls=cls))

            blocks.append(_generate_visit_for_class(cls=cls))

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    writer.write(
        f"""\
/// <summary>
/// Serialize recursively the instances as XML elements.
/// </summary>
internal class VisitorWithWriter
{I}: Visitation.AbstractVisitorWithContext<Xml.XmlWriter>
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // internal class VisitorWithWriter")

    return Stripped(writer.getvalue()), None


def _generate_serialize(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the static serializer."""
    blocks = [
        Stripped(
            f"""\
[CodeAnalysis.SuppressMessage("ReSharper", "InconsistentNaming")]
private static readonly VisitorWithWriter _visitorWithWriter = (
{I}new VisitorWithWriter());"""
        ),
        Stripped(
            f"""\
/// <summary>
/// Serialize an instance of the meta-model to XML.
/// </summary>
public static void To(
{I}Aas.IClass that,
{I}Xml.XmlWriter writer)
{{
{I}Serialize._visitorWithWriter.Visit(
{II}that, writer);
}}"""
        ),
    ]  # type: List[Stripped]

    writer = io.StringIO()
    writer.write(
        """\
/// <summary>
/// Serialize instances of meta-model classes to XML.
/// </summary>
"""
    )

    first_cls = (
        symbol_table.classes[0] if len(symbol_table.classes) > 0 else None
    )  # type: Optional[intermediate.ClassUnion]

    if first_cls is not None:
        cls_name: str
        if isinstance(first_cls, intermediate.AbstractClass):
            cls_name = csharp_naming.interface_name(first_cls.name)
        elif isinstance(first_cls, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(first_cls.name)
        else:
            assert_never(first_cls)

        an_instance_variable = csharp_naming.variable_name(Identifier("an_instance"))

        writer.write(
            f"""\
/// <example>
/// Here is an example how to serialize an instance of {cls_name}:
/// <code>
/// var {an_instance_variable} = new Aas.{cls_name}(
///     /* ... some constructor arguments ... */
/// );
/// var writer = new System.Xml.XmlWriter( /* some arguments */ );
/// Serialize.To(
/// {I}{an_instance_variable},
/// {I}writer);
/// </code>
/// </example>
"""
        )

    writer.write(
        """\
public static class Serialize
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public static class Serialize")

    return Stripped(writer.getvalue())


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    namespace: csharp_common.NamespaceIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the C# code for the general serialization.

    The ``namespace`` defines the AAS C# namespace.
    """
    xmlization_blocks = []  # type: List[Stripped]

    errors = []  # type: List[Error]

    deserialize_impl_block, deserialize_impl_errors = _generate_deserialize_impl(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if deserialize_impl_errors is not None:
        errors.extend(deserialize_impl_errors)
    else:
        assert deserialize_impl_block is not None
        xmlization_blocks.append(deserialize_impl_block)

    xmlization_blocks.append(
        Stripped(
            f"""\
/// <summary>
/// Represent a critical error during the deserialization.
/// </summary>
public class Exception : System.Exception
{{
{I}public readonly string Path;
{I}public readonly string Cause;
{I}public Exception(string path, string cause)
{II}: base($"{{cause}} at: {{(path == "" ? "the beginning" : path)}}")
{I}{{
{II}Path = path;
{II}Cause = cause;
{I}}}
}}"""
        )
    )

    xmlization_blocks.append(_generate_deserialize(symbol_table=symbol_table))

    visitor_block, visitor_errors = _generate_visitor(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if visitor_errors is not None:
        errors.extend(visitor_errors)
    else:
        assert visitor_block is not None
        xmlization_blocks.append(visitor_block)

    if len(errors) > 0:
        return None, errors

    xmlization_blocks.append(_generate_serialize(symbol_table=symbol_table))

    xmlization_writer = io.StringIO()

    xml_namespace_literal = csharp_common.string_literal(
        symbol_table.meta_model.xml_namespace
    )

    xmlization_writer.write(
        f"""\
namespace {namespace}
{{
{I}/// <summary>
{I}/// Provide de/serialization of meta-model classes to/from XML.
{I}/// </summary>
{I}public static class Xmlization
{I}{{
{II}/// The XML namespace of the meta-model
{II}[CodeAnalysis.SuppressMessage("ReSharper", "InconsistentNaming")]
{II}public static readonly string NS = (
{III}{xml_namespace_literal});

"""
    )

    for i, xmlization_block in enumerate(xmlization_blocks):
        if i > 0:
            xmlization_writer.write("\n\n")

        xmlization_writer.write(textwrap.indent(xmlization_block, II))

    xmlization_writer.write(f"\n{I}}}  // public static class Xmlization")
    xmlization_writer.write(f"\n}}  // namespace {namespace}")

    using_directives = []  # type: List[Stripped]
    using_directives.extend(
        csharp_common.generate_using_aas_directive_if_necessary(namespace)
    )

    using_directives.append(
        Stripped(
            """\
using CodeAnalysis = System.Diagnostics.CodeAnalysis;
using Xml = System.Xml;

using System.Collections.Generic;  // can't alias"""
        )
    )

    # pylint: disable=line-too-long
    blocks = [
        csharp_common.WARNING,
        Stripped("\n".join(using_directives)),
        Stripped(xmlization_writer.getvalue()),
        csharp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        assert not block.startswith("\n")
        assert not block.endswith("\n")
        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None



"""Generate the code to test ``descend_once`` methods."""

import io

from icontract import ensure

from aas_core_codegen.common import Stripped
from aas_core_codegen.python import common as python_common
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


@ensure(
    lambda result: result.endswith("\n"),
    "Trailing newline mandatory for valid end-of-files",
)
def generate(qualified_module_name: python_common.QualifiedModuleName) -> str:
    """
    Generate the code to unit test ``descend_once`` methods..

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    blocks = [
        Stripped(
            f'''\
"""Test :py:method:`{qualified_module_name}.types.Class.descend_once`."""'''
        ),
        python_common.WARNING,
        Stripped(
            """\
# pylint: disable=missing-docstring"""
        ),
        Stripped(
            """\
import unittest"""
        ),
        Stripped(
            """\
import tests.common
import tests.common_xmlization"""
        ),
        Stripped(
            f"""\
class TestWithExpectedMaximal(unittest.TestCase):
{I}def test_descend_once_against_recorded_trace_log(self) -> None:
{II}xml_expected_dir = tests.common.TEST_DATA_DIR / "Xml" / "Expected"

{II}for path in sorted(xml_expected_dir.glob("**/maximal.xml")):
{III}instance = tests.common_xmlization.must_load(path)

{III}rel_path = path.relative_to(xml_expected_dir)

{III}expected_path = (
{IIII}tests.common.TEST_DATA_DIR
{IIII}/ "descend_once"
{IIII}/ rel_path.parent
{IIII}/ f"{{rel_path.name}}.trace"
{III})

{III}log = [
{IIII}tests.common.trace(something)
{IIII}for something in instance.descend_once()
{III}]

{III}got_text = tests.common.trace_log_as_text_file_content(log)

{III}tests.common.record_or_check(expected_path, got_text)"""
        ),
        Stripped(
            f"""\
if __name__ == "__main__":
{I}unittest.main()"""
        ),
        python_common.WARNING,
    ]

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue()



"""Generate the test code for the de/serialization of instances in XML."""

from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for the de/serialization of instances in XML.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    xml_namespace_literal = csharp_common.string_literal(
        symbol_table.meta_model.xml_namespace
    )

    blocks = [
        Stripped(
            f"""\
private static void CheckElementsEqual(
{I}XElement expected,
{I}XElement got,
{I}out Reporting.Error? error)
{{
{I}error = null;

{I}if (expected.Name.LocalName != got.Name.LocalName)
{I}{{
{II}error = new Reporting.Error(
{III}"Mismatch in element names: " +
{III}$"{{expected}} != {{got}}"
{II});
{II}return;
{I}}}

{I}string? expectedContent = (expected.FirstNode as XText)?.Value;
{I}string? gotContent = (got.FirstNode as XText)?.Value;

{I}if (expectedContent != gotContent)
{I}{{
{II}error = new Reporting.Error(
{III}$"Mismatch in element contents: {{expected}} != {{got}}"
{II});
{II}return;
{I}}}

{I}var expectedChildren = expected.Elements().ToList();
{I}var gotChildren = got.Elements().ToList();

{I}if (expectedChildren.Count != gotChildren.Count)
{I}{{
{II}error = new Reporting.Error(
{III}$"Mismatch in child elements: {{expected}} != {{got}}"
{II});
{II}return;
{I}}}

{I}for (int i = 0; i < expectedChildren.Count; i++)
{I}{{
{II}CheckElementsEqual(
{III}expectedChildren[i],
{III}gotChildren[i],
{III}out error);

{II}if (error != null)
{II}{{
{III}error.PrependSegment(
{IIII}new Reporting.IndexSegment(i));

{III}error.PrependSegment(
{IIII}new Reporting.NameSegment(
{IIIII}expected.Name.ToString()));
{II}}}
{I}}}
}}"""
        ),
        Stripped(
            f"""\
private static void AssertSerializeDeserializeEqualsOriginal(
{I}Aas.IClass instance, string path)
{{
{I}// Serialize
{I}var outputBuilder = new System.Text.StringBuilder();

{I}{{
{II}using var writer = System.Xml.XmlWriter.Create(
{III}outputBuilder,
{III}new System.Xml.XmlWriterSettings()
{III}{{
{IIII}Encoding = System.Text.Encoding.UTF8,
{IIII}OmitXmlDeclaration = true
{III}}}
{II});
{II}Aas.Xmlization.Serialize.To(
{III}instance,
{III}writer);
{I}}}

{I}string outputText = outputBuilder.ToString();

{I}// Compare input == output
{I}{{
{II}using var outputReader = new System.IO.StringReader(outputText);
{II}var gotDoc = XDocument.Load(outputReader);

{II}Assert.AreEqual(
{III}gotDoc.Root?.Name.Namespace.ToString(),
{III}{xml_namespace_literal});

{II}foreach (var child in gotDoc.Descendants())
{II}{{
{III}Assert.AreEqual(
{IIII}child.GetDefaultNamespace().NamespaceName,
{IIII}{xml_namespace_literal});
{II}}}

{II}var expectedDoc = XDocument.Load(path);

{II}CheckElementsEqual(
{III}expectedDoc.Root!,
{III}gotDoc.Root!,
{III}out Reporting.Error? inequalityError);

{II}if (inequalityError != null)
{II}{{
{III}Assert.Fail(
{IIII}$"The original XML from {{path}} is unequal the serialized XML: " +
{IIII}$"#/{{Reporting.GenerateRelativeXPath(inequalityError.PathSegments)}}: " +
{IIII}inequalityError.Cause
{III});
{II}}}
{I}}}
}}"""
        ),
        Stripped(
            f"""\
private static void AssertEqualsExpectedOrRerecordDeserializationException(
{I}Aas.Xmlization.Exception? exception,
{I}string path)
{{
{I}if (exception == null)
{I}{{
{II}Assert.Fail(
{III}$"Expected a Xmlization exception when de-serializing {{path}}, but got none."
{II});
{I}}}
{I}else
{I}{{
{II}string exceptionPath = path + ".exception";
{II}string got = exception.Message;
{II}if (Aas.Tests.Common.RecordMode)
{II}{{
{III}System.IO.File.WriteAllText(exceptionPath, got);
{II}}}
{II}else
{II}{{
{III}if (!System.IO.File.Exists(exceptionPath))
{III}{{
{IIII}throw new System.IO.FileNotFoundException(
{IIIII}"The file with the recorded exception does not " +
{IIIII}$"exist: {{exceptionPath}}; maybe you want to set the environment " +
{IIIII}$"variable {{Aas.Tests.Common.RecordModeEnvironmentVariableName}}?");
{III}}}

{III}string expected = System.IO.File.ReadAllText(exceptionPath);
{III}Assert.AreEqual(
{IIII}expected.Replace("\\r\\n", "\\n"),
{IIII}got.Replace("\\r\\n", "\\n"),
{IIII}$"The expected exception does not match the actual one for the file {{path}}");
{II}}}
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_csharp = csharp_naming.class_name(concrete_cls.name)
        cls_name_xml = naming.xml_class_name(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{cls_name_csharp}_ok()
{{
{I}var paths = Directory.GetFiles(
{II}Path.Combine(
{III}Aas.Tests.Common.TestDataDir,
{III}"Xml",
{III}"Expected",
{III}{csharp_common.string_literal(cls_name_xml)}
{II}),
{II}"*.xml",
{II}System.IO.SearchOption.AllDirectories).ToList();
{I}paths.Sort();

{I}foreach (var path in paths)
{I}{{
{II}using var xmlReader = System.Xml.XmlReader.Create(path);

{II}var instance = Aas.Xmlization.Deserialize.{cls_name_csharp}From(
{III}xmlReader);

{II}var errors = Aas.Verification.Verify(instance).ToList();
{II}Aas.Tests.Common.AssertNoVerificationErrors(errors, path);

{II}AssertSerializeDeserializeEqualsOriginal(
{III}instance, path);
{I}}}
}}  // public void Test_{cls_name_csharp}_ok"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{cls_name_csharp}_deserialization_fail()
{{
{I}foreach (
{II}string causeDir in
{II}Directory.GetDirectories(
{III}Path.Combine(
{IIII}Aas.Tests.Common.TestDataDir,
{IIII}"Xml",
{IIII}"Unexpected",
{IIII}"Unserializable"
{III})
{II})
{I})
{I}{{
{II}string clsDir = Path.Combine(
{III}causeDir,
{III}{csharp_common.string_literal(cls_name_xml)}
{II});

{II}if (!Directory.Exists(clsDir))
{II}{{
{III}// No examples of {cls_name_csharp} for the failure cause.
{III}continue;
{II}}}

{II}var paths = Directory.GetFiles(
{III}clsDir,
{III}"*.xml",
{III}System.IO.SearchOption.AllDirectories).ToList();
{II}paths.Sort();

{II}foreach (var path in paths)
{II}{{
{III}using var xmlReader = System.Xml.XmlReader.Create(path);

{III}Aas.Xmlization.Exception? exception = null;

{III}try
{III}{{
{IIII}_ = Aas.Xmlization.Deserialize.{cls_name_csharp}From(
{IIIII}xmlReader);
{III}}}
{III}catch (Aas.Xmlization.Exception observedException)
{III}{{
{IIII}exception = observedException;
{III}}}

{III}AssertEqualsExpectedOrRerecordDeserializationException(
{IIII}exception, path);
{II}}}
{I}}}
}}  // public void Test_{cls_name_csharp}_deserialization_fail"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_{cls_name_csharp}_verification_fail()
{{
{I}foreach (
{II}string causeDir in
{II}Directory.GetDirectories(
{III}Path.Combine(
{IIII}Aas.Tests.Common.TestDataDir,
{IIII}"Xml",
{IIII}"Unexpected",
{IIII}"Invalid"
{III})
{II})
{I})
{I}{{
{II}string clsDir = Path.Combine(
{III}causeDir,
{III}{csharp_common.string_literal(cls_name_xml)}
{II});

{II}if (!Directory.Exists(clsDir))
{II}{{
{III}// No examples of {cls_name_csharp} for the failure cause.
{III}continue;
{II}}}

{II}var paths = Directory.GetFiles(
{III}clsDir,
{III}"*.xml",
{III}System.IO.SearchOption.AllDirectories).ToList();
{II}paths.Sort();

{II}foreach (var path in paths)
{II}{{
{III}using var xmlReader = System.Xml.XmlReader.Create(path);

{III}var instance = Aas.Xmlization.Deserialize.{cls_name_csharp}From(
{IIII}xmlReader);

{III}var errors = Aas.Verification.Verify(instance).ToList();
{III}Aas.Tests.Common.AssertEqualsExpectedOrRerecordVerificationErrors(
{IIII}errors, path);
{II}}}
{I}}}
}}  // public void Test_{cls_name_csharp}_verification_fail"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using Directory = System.IO.Directory;
using Path = System.IO.Path;

using NUnit.Framework; // can't alias
using System.Linq;  // can't alias
using System.Xml.Linq; // can't alias

namespace {namespace}.Tests
{{
{I}public class TestXmlizationOfConcreteClasses
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestXmlizationOfConcreteClasses
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Fix regular expressions for different engines."""
from typing import Tuple, List

from icontract import require, ensure

from aas_core_codegen.parse.retree import (
    _visitor as retree_visitor,
    _types as retree_types,
)


class _FixForUTF16Regex(retree_visitor.PassThroughVisitor):
    """
    Modify the pattern in-place so that UTF-32 can be dealt by UTF-16-only regex engine.

    Specifically, we need to split UTF-32 characters in the two surrogates since
    UTF-16-only regex engines only work with UTF-16.

    The characters in the range ``\\uD800`` to ``\\uDFFF`` are reserved, so they can
    be used as high surrogates. However, it is still valid to use them as standalone
    characters. Hence, if you use them in your pattern, and the pattern also involves
    UTF-32 characters, the transformed pattern might be more permissive than
    the original one.

    For the problem, see:

    * https://stackoverflow.com/questions/364009/c-sharp-regular-expressions-with-uxxxxxxxx-characters-in-the-pattern
    * https://stackoverflow.com/questions/47605037/unicode-character-range-not-being-consumed-by-regex

    For the UTF-16 specs, ranges and computation of low/high surrogates, see:

    * https://en.wikipedia.org/wiki/UTF-16#Description
    * http://www.russellcottrell.com/greek/utilities/surrogatepaircalculator.htm
    """

    _SUPPLEMENTARY_PLANE_START = 0x00010000
    _SUPPLEMENTARY_PLANE_END = 0x0010FFFF

    # fmt: off
    @staticmethod
    @require(
        lambda code:
        _FixForUTF16Regex._SUPPLEMENTARY_PLANE_START
        <= code <=
        _FixForUTF16Regex._SUPPLEMENTARY_PLANE_END

    )
    @ensure(
        lambda result:
        0xD800 <= result[0] <= 0xDBFF
        and 0xDC00 <= result[1] <= 0xDFFF,
        "High and low surrogates in the expected range"
    )
    # fmt: on
    def _convert_to_surrogates(code: int) -> Tuple[int, int]:
        """Convert a UTF-32 character into its surrogates."""
        high_surrogate = (code - 0x10000) // 0x400 + 0xD800
        low_surrogate = (code - 0x10000) % 0x400 + 0xDC00

        return high_surrogate, low_surrogate

    @staticmethod
    @require(lambda term: isinstance(term.value, retree_types.Char))
    def _character_literal_to_surrogates_if_necessary(
        term: retree_types.Term,
    ) -> List[retree_types.Term]:
        """Expand the character literal to two surrogate characters if necessary."""
        # NOTE (mristin, 2022-06-10):
        # This assertion is needed for mypy.
        assert isinstance(term.value, retree_types.Char)

        if ord(term.value.character) < _FixForUTF16Regex._SUPPLEMENTARY_PLANE_START:
            return [term]

        high_surrogate, low_surrogate = _FixForUTF16Regex._convert_to_surrogates(
            code=ord(term.value.character)
        )

        # NOTE (mristin, 2022-06-10):
        # We explicitly encode the character since otherwise this split can not be
        # traced back meaningfully.
        high_surrogate_char = retree_types.Char(
            character=chr(high_surrogate), explicitly_encoded=True
        )
        low_surrogate_char = retree_types.Char(
            character=chr(low_surrogate), explicitly_encoded=True
        )

        output = []  # type: List[retree_types.Term]

        if term.quantifier is not None:
            # NOTE (mristin, 2022-06-10):
            # We need to put the surrogates in a group so that the quantifier
            # applies to both.
            value = retree_types.Group(
                union=retree_types.UnionExpr(
                    uniates=[
                        retree_types.Concatenation(
                            concatenants=[
                                retree_types.Term(
                                    value=high_surrogate_char, quantifier=None
                                ),
                                retree_types.Term(
                                    value=low_surrogate_char, quantifier=None
                                ),
                            ]
                        )
                    ]
                )
            )

            output.append(retree_types.Term(value=value, quantifier=term.quantifier))
        else:
            # NOTE (mristin, 2022-06-10):
            # When there is no quantifier, we can simply inject the surrogates
            # as character literals.
            output.append(retree_types.Term(value=high_surrogate_char, quantifier=None))
            output.append(retree_types.Term(value=low_surrogate_char, quantifier=None))

        return output

    @staticmethod
    def _produce_char_char(
        first_code: int, second_code: int
    ) -> retree_types.Concatenation:
        """
        Produce a concatenation as ``{u(code)}{u(code)}``.

        This method helped us to make
        :method:`~_expand_char_set_to_surrogates_if_necessary` somewhat manageable since
        we struggled with the complexity of its code.
        """
        return retree_types.Concatenation(
            concatenants=[
                retree_types.Term(
                    value=retree_types.Char(
                        character=chr(first_code), explicitly_encoded=True
                    ),
                    quantifier=None,
                ),
                retree_types.Term(
                    value=retree_types.Char(
                        character=chr(second_code), explicitly_encoded=True
                    ),
                    quantifier=None,
                ),
            ]
        )

    @staticmethod
    def _produce_char_char_set(
        code: int, range_start: int, range_end: int
    ) -> retree_types.Concatenation:
        """
        Produce a concatenation as ``{u(code)}[{u(range_start)-u(range_end)}]``.

        This method helped us to make
        :method:`~_expand_char_set_to_surrogates_if_necessary` somewhat manageable since
        we struggled with the complexity of its code.
        """
        return retree_types.Concatenation(
            concatenants=[
                retree_types.Term(
                    value=retree_types.Char(
                        character=chr(code), explicitly_encoded=True
                    ),
                    quantifier=None,
                ),
                retree_types.Term(
                    value=retree_types.CharSet(
                        complementing=False,
                        ranges=[
                            retree_types.Range(
                                start=retree_types.Char(
                                    character=chr(range_start), explicitly_encoded=True
                                ),
                                end=retree_types.Char(
                                    character=chr(range_end), explicitly_encoded=True
                                ),
                            )
                        ],
                    ),
                    quantifier=None,
                ),
            ]
        )

    @staticmethod
    def _produce_char_set_char_set(
        first_range_start: int,
        first_range_end: int,
        second_range_start: int,
        second_range_end: int,
    ) -> retree_types.Concatenation:
        """
        Produce a concatenation of the character sets.

        This method helped us to make
        :method:`~_expand_char_set_to_surrogates_if_necessary` somewhat manageable since
        we struggled with the complexity of its code.
        """
        return retree_types.Concatenation(
            concatenants=[
                retree_types.Term(
                    value=retree_types.CharSet(
                        complementing=False,
                        ranges=[
                            retree_types.Range(
                                start=retree_types.Char(
                                    character=chr(first_range_start),
                                    explicitly_encoded=True,
                                ),
                                end=retree_types.Char(
                                    character=chr(first_range_end),
                                    explicitly_encoded=True,
                                ),
                            )
                        ],
                    ),
                    quantifier=None,
                ),
                retree_types.Term(
                    value=retree_types.CharSet(
                        complementing=False,
                        ranges=[
                            retree_types.Range(
                                start=retree_types.Char(
                                    character=chr(second_range_start),
                                    explicitly_encoded=True,
                                ),
                                end=retree_types.Char(
                                    character=chr(second_range_end),
                                    explicitly_encoded=True,
                                ),
                            )
                        ],
                    ),
                    quantifier=None,
                ),
            ]
        )

    @staticmethod
    @require(lambda term: isinstance(term.value, retree_types.CharSet))
    def _expand_char_set_to_surrogates_if_necessary(
        term: retree_types.Term,
    ) -> List[retree_types.Term]:
        """Extend the set with surrogates or expand it to a union, if necessary."""
        ranges_wo_utf32 = []  # type: List[retree_types.Range]
        ranges_w_utf32 = []  # type: List[retree_types.Range]

        # Needed for mypy
        assert isinstance(term.value, retree_types.CharSet)

        for a_range in term.value.ranges:
            if (
                ord(a_range.start.character)
                < _FixForUTF16Regex._SUPPLEMENTARY_PLANE_START
            ) and (
                a_range.end is None
                or (
                    ord(a_range.end.character)
                    < _FixForUTF16Regex._SUPPLEMENTARY_PLANE_START
                )
            ):
                ranges_wo_utf32.append(a_range)
            else:
                ranges_w_utf32.append(a_range)

        if len(ranges_w_utf32) == 0:
            return [term]

        assert not term.value.complementing, (
            "Complementing character sets with one or more ranges "
            "involving UTF-32 characters can not be supported, "
            "since we can not represent them relying "
            "on an UTF-16-only regex engine.\n"
            "\n"
            "This should have been detected before and returned as an error.",
        )

        # NOTE (mristin, 2022-06-11):
        # Expand the character set into a union of:
        # 1) A character set without any UTF-32 characters
        # 2) One or more character sets representing the ranges involving
        #    UTF-32 characters.

        # noinspection PyListCreation
        uniates = []  # type: List[retree_types.Concatenation]

        if len(ranges_wo_utf32) > 0:
            uniates.append(
                retree_types.Concatenation(
                    concatenants=[
                        retree_types.Term(
                            value=retree_types.CharSet(
                                complementing=False, ranges=ranges_wo_utf32
                            ),
                            quantifier=None,
                        )
                    ]
                )
            )

        for a_range in ranges_w_utf32:
            high_start, low_start = _FixForUTF16Regex._convert_to_surrogates(
                code=ord(a_range.start.character)
            )

            if (a_range.end is None) or (
                a_range.start.character == a_range.end.character
            ):
                uniates.append(
                    _FixForUTF16Regex._produce_char_char(
                        first_code=high_start, second_code=low_start
                    )
                )
            else:
                assert ord(a_range.start.character) < ord(a_range.end.character), (
                    "Start of a range must be smaller than its end.\n"
                    "\n"
                    "This is expected to be detected at the parse stage."
                )

                high_end, low_end = _FixForUTF16Regex._convert_to_surrogates(
                    code=ord(a_range.end.character)
                )

                # noinspection SpellCheckingInspection
                if high_start == high_end:
                    # {high start}[{low start}-{low end}]
                    uniates.append(
                        _FixForUTF16Regex._produce_char_char_set(
                            code=high_start, range_start=low_start, range_end=low_end
                        )
                    )
                else:
                    # {high start}[{low start}-\uDFFF]
                    uniates.append(
                        _FixForUTF16Regex._produce_char_char_set(
                            code=high_start, range_start=low_start, range_end=0xDFFF
                        )
                    )

                    if high_end - high_start > 1:
                        # noinspection SpellCheckingInspection
                        if high_start + 1 == high_end - 1:
                            # {high start + 1}[\uDC00-\uDFFF]
                            uniates.append(
                                _FixForUTF16Regex._produce_char_char_set(
                                    code=high_start + 1,
                                    range_start=0xDC00,
                                    range_end=0xDFFF,
                                )
                            )
                        else:
                            # [{high start + 1}-{high end - 1}][\uDC00-\uDFFF]
                            uniates.append(
                                _FixForUTF16Regex._produce_char_set_char_set(
                                    first_range_start=high_start + 1,
                                    first_range_end=high_end - 1,
                                    second_range_start=0xDC00,
                                    second_range_end=0xDFFF,
                                )
                            )

                    uniates.append(
                        _FixForUTF16Regex._produce_char_char_set(
                            code=high_end, range_start=0xDC00, range_end=low_end
                        )
                    )

        return [
            retree_types.Term(
                value=retree_types.Group(union=retree_types.UnionExpr(uniates=uniates)),
                quantifier=term.quantifier,
            )
        ]

    def visit_concatenation(self, node: retree_types.Concatenation) -> None:
        """Convert character literals to UTF-16, where appropriate."""
        new_concatenants = []  # type: List[retree_types.Term]

        for concatenant in node.concatenants:
            if isinstance(concatenant.value, retree_types.Char):
                new_concatenants.extend(
                    _FixForUTF16Regex._character_literal_to_surrogates_if_necessary(
                        term=concatenant
                    )
                )
            elif isinstance(concatenant.value, retree_types.CharSet):
                new_concatenants.extend(
                    _FixForUTF16Regex._expand_char_set_to_surrogates_if_necessary(
                        term=concatenant
                    )
                )
            else:
                new_concatenants.append(concatenant)

        node.concatenants = new_concatenants
        for concatenant in new_concatenants:
            self.visit(concatenant)


_FIX_FOR_UTF16_REGEX = _FixForUTF16Regex()


def fix_for_utf16_regex_in_place(regex: retree_types.Regex) -> None:
    """
    Modify the pattern in-place so that UTF-32 can be dealt by UTF-16-only regex engine.

    Specifically, we need to split UTF-32 characters in the two surrogates since
    UTF-16-only regex engines only work with UTF-16.

    The characters in the range ``\\uD800`` to ``\\uDFFF`` are reserved, so they can
    be used as high surrogates. However, it is still valid to use them as standalone
    characters. Hence, if you use them in your pattern, and the pattern also involves
    UTF-32 characters, the transformed pattern might be more permissive than
    the original one.

    For the problem, see:

    * https://stackoverflow.com/questions/364009/c-sharp-regular-expressions-with-uxxxxxxxx-characters-in-the-pattern
    * https://stackoverflow.com/questions/47605037/unicode-character-range-not-being-consumed-by-regex

    For the UTF-16 specs, ranges and computation of low/high surrogates, see:

    * https://en.wikipedia.org/wiki/UTF-16#Description
    * http://www.russellcottrell.com/greek/utilities/surrogatepaircalculator.htm
    """
    _FIX_FOR_UTF16_REGEX.visit(regex)


assert fix_for_utf16_regex_in_place.__doc__ == _FIX_FOR_UTF16_REGEX.__doc__



"""Handle implementation snippets regardless for all implementation languages."""

import pathlib
import re
from typing import cast, Mapping, Tuple, Optional, List

from icontract import require, ensure

from aas_core_codegen.common import Stripped

# noinspection RegExpSimplifiable
IMPLEMENTATION_KEY_RE = re.compile("[a-zA-Z_][a-zA-Z_0-9.]*(/[a-zA-Z_][a-zA-Z_0-9.]*)*")


class ImplementationKey(str):
    """Represent a key in the map of specific implementations."""

    @require(lambda key: IMPLEMENTATION_KEY_RE.fullmatch(key))
    def __new__(cls, key: str) -> "ImplementationKey":
        return cast(ImplementationKey, key)


SpecificImplementations = Mapping[ImplementationKey, Stripped]


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def read_from_directory(
    snippets_dir: pathlib.Path,
) -> Tuple[Optional[SpecificImplementations], Optional[List[str]]]:
    """
    Read all the implementation-specific code snippets from the ``snippets_dir``.

    :return: either the map of the implementations, or the errors
    """
    mapping = dict()  # pylint: disable=use-dict-literal

    errors = []  # type: List[str]
    for pth in snippets_dir.glob("**/*"):
        # NOTE (mristin, 2022-08-25):
        # Ignore hidden or special files. In particular, we do not want Git-related
        # files such as ``.gitignore`` to be included as snippets.
        if pth.name.startswith("."):
            continue

        if pth.is_dir():
            continue

        maybe_key = (pth.relative_to(snippets_dir).parent / pth.name).as_posix()
        if IMPLEMENTATION_KEY_RE.fullmatch(maybe_key) is None:
            errors.append(
                f"The snippet key is not valid "
                f"according to {IMPLEMENTATION_KEY_RE.pattern}: {maybe_key}"
            )
            continue

        key = ImplementationKey(maybe_key)

        try:
            value = Stripped(pth.read_text(encoding="utf-8").strip())
        except UnicodeDecodeError as error:
            errors.append(
                f"The snippet file is not a valid UTF-8: {pth}. "
                f"This was the decoding error: {error}"
            )
            continue

        mapping[key] = value

    if errors:
        return None, errors

    return mapping, None



"""Generate the common functions to de/serialize instances of a class."""


import io
import textwrap
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import INDENT as I, INDENT2 as II


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier, symbol_table: intermediate.SymbolTable
) -> str:
    """
    Generate the common functions to de/serialize instances of a class.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = []  # type: List[str]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_csharp = csharp_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
public static Aas.{cls_name_csharp} LoadMaximal{cls_name_csharp}()
{{
{I}string path = Path.Combine(
{II}Aas.Tests.Common.TestDataDir,
{II}"Json",
{II}"Expected",
{II}{csharp_common.string_literal(cls_name_json)},
{II}"maximal.json");

{I}var node = Aas.Tests.CommonJson.ReadFromFile(path);

{I}var instance = Aas.Jsonization.Deserialize.{cls_name_csharp}From(
{II}node);

{I}return instance;
}}  // public static Aas.{cls_name_csharp} LoadMaximal{cls_name_csharp}

public static Aas.{cls_name_csharp} LoadMinimal{cls_name_csharp}()
{{
{I}string path = Path.Combine(
{II}Aas.Tests.Common.TestDataDir,
{II}"Json",
{II}"Expected",
{II}{csharp_common.string_literal(cls_name_json)},
{II}"minimal.json");

{I}var node = Aas.Tests.CommonJson.ReadFromFile(path);

{I}var instance = Aas.Jsonization.Deserialize.{cls_name_csharp}From(
{II}node);

{I}return instance;
}}  // public static Aas.{cls_name_csharp} LoadMinimal{cls_name_csharp}"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using Path = System.IO.Path;

namespace {namespace}.Tests
{{
{I}/// <summary>
{I}/// Provide methods to load instances from JSON test data.
{I}/// </summary>
{I}public static class CommonJsonization
{I}{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, II))

    writer.write(
        f"""
{I}}}  // class CommonJsonization
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""
    )

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate Java code for enhancing model classes."""

import io
import textwrap
from typing import Tuple, Optional, List

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations
from aas_core_codegen.common import (
    Error,
    Identifier,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.java import (
    common as java_common,
    naming as java_naming,
)
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def _generate_delegate_method(method: intermediate.Method) -> Stripped:
    """Generate the delegated method to ``instance``."""
    returns = (
        java_common.generate_type(method.returns)
        if method.returns is not None
        else "void"
    )

    arg_types_names = [
        (
            java_common.generate_type(arg.type_annotation),
            java_naming.argument_name(arg.name),
        )
        for arg in method.arguments
    ]

    method_name = java_naming.method_name(method.name)

    return_prefix = "return " if method.returns is not None else ""

    if len(method.arguments) == 0:
        return Stripped(
            f"""\
public {returns} {method_name}() {{
{I}{return_prefix}instance.{method_name}();
}}"""
        )

    arguments_definition = ",\n".join(
        f"{arg_type} {arg_name}" for arg_type, arg_name in arg_types_names
    )

    arguments_delegation = ",\n".join(arg_name for _, arg_name in arg_types_names)

    return Stripped(
        f"""\
public {returns} {method_name}(
{I}{indent_but_first_line(arguments_definition, I)}
)
{{
{I}{return_prefix}instance.{method_name}(
{II}{indent_but_first_line(arguments_delegation, II)}
{I});
}}"""
    )


def _generate_enhanced_abstract_class(
    package: java_common.PackageIdentifier,
) -> java_common.JavaFile:
    enhanced = Stripped(
        f"""\
public abstract class Enhanced<EnhancementT> {{
{I}protected final EnhancementT enhancement;

{I}protected Enhanced(EnhancementT enhancement) {{
{II}this.enhancement = enhancement;
{I}}}

{I}EnhancementT getEnhancement() {{
{II}return enhancement;
{I}}}
}}"""
    )

    blocks = [
        java_common.WARNING,
        Stripped(f"package {package}.enhancing;"),
        enhanced,
        java_common.WARNING,
    ]  # type: List[Stripped]

    code = "\n\n".join(blocks)

    return java_common.JavaFile(
        Stripped("Enhanced.java"),
        f"{code}\n",
    )


def _generate_unwrapper_class(
    package: java_common.PackageIdentifier,
) -> java_common.JavaFile:
    imports = [
        Stripped("import java.util.Optional;"),
        Stripped(f"import {package}.types.model.*;"),
    ]  # type: List[Stripped]

    unwrapper = Stripped(
        f"""\
/**
 * Unwrap enhancements from the wrapped instances.
 *
 * @param <EnhancementT> structure of the expected enhancement
 */
public class Unwrapper<EnhancementT> {{
{I}/**
{I} * Unwrap the given model instance.
{I} *
{I} * @param that model instance to be unwrapped
{I} * @return Enhancement, or {{@link java.util.Optional#empty()}} if {{@code that}}
{I} * has not been wrapped yet.
{I} */
{I}public Optional<EnhancementT> unwrap(IClass that)
{I}{{
{II}if (that instanceof Enhanced) {{
{III}@SuppressWarnings("unchecked")
{III}Enhanced<EnhancementT> enhanced = (Enhanced<EnhancementT>) that;
{III}return Optional.of(enhanced.getEnhancement());
{II}}} else {{
{III}return Optional.empty();
{II}}}
{I}}}

{I}/**
{I} * Unwrap the given model instance.
{I} *
{I} * @param that model instance to be unwrapped
{I} * @return Enhancement wrapped around {{@code that}}
{I} */
{I}public EnhancementT mustUnwrap(IClass that)
{I}{{
{II}Optional<EnhancementT> value = unwrap(that);
{II}if (!value.isPresent()) {{
{III}throw new IllegalArgumentException(
{IIII}"Expected the instance to have been wrapped, but it was not: " + that
{III});
{II}}}
{II}return value.get();
{I}}}
}}"""
    )

    blocks = [
        java_common.WARNING,
        Stripped(f"package {package}.enhancing;"),
        Stripped("\n".join(imports)),
        unwrapper,
        java_common.WARNING,
    ]  # type: List[Stripped]

    code = "\n\n".join(blocks)

    return java_common.JavaFile(
        Stripped("Unwrapper.java"),
        f"{code}\n",
    )


def _generate_enhancer_class(
    package: java_common.PackageIdentifier,
) -> java_common.JavaFile:
    imports = [
        Stripped("import java.util.function.Function;"),
        Stripped("import java.util.Optional;"),
        Stripped(f"import {package}.enhancing.Unwrapper;"),
        Stripped(f"import {package}.types.model.*;"),
    ]  # type: List[Stripped]

    enhancer = Stripped(
        f"""\
/**
 * Wrap and unwrap the instances of model classes with enhancement.
 *
 * @param <EnhancementT> structure of the enhancement
 */
public class Enhancer<EnhancementT> extends Unwrapper<EnhancementT> {{
{I}private final Wrapper<EnhancementT> wrapper;

{I}/**
{I} * @param enhancementFactory how to enhance the instances.
{I} *
{I} * <p>If it returns {{@code null}}, the instance will not be wrapped. However,
{I} * the wrapping will continue recursively.
{I} */
{I}public Enhancer(
{II}Function<IClass, Optional<EnhancementT>> enhancementFactory
{I}) {{
{II}this.wrapper = new Wrapper<>(enhancementFactory);
{I}}}

{I}/**
{I} * Wrap the instance with an enhancement.
{I} *
{I} * <p>Double wraps are not allowed to prevent runtime leakage.
{I} *
{I} * <p>If you use references to the instance objects, you have to update them
{I} * after the wrapping, as the wrapping is recursive.
{I} *
{I} * @param that model instance to be wrapped
{I} * @return {{@code that}} instance wrapped recursively with enhancements
{I} */
{I}public IClass wrap(
{II}IClass that
{I}) {{
{II}IClass wrapped;
{II}try {{
{III}wrapped = wrapper.transform(that);
{II}}} catch (IllegalArgumentException exception) {{
{III}throw new UnsupportedOperationException(
{IIII}"Expected the wrapped instance to be an instance of IClass, " +
{IIII}"but got: " + that
{III});
{II}}}

{II}return wrapped;
{I}}}
}}"""
    )

    blocks = [
        java_common.WARNING,
        Stripped(f"package {package}.enhancing;"),
        Stripped("\n".join(imports)),
        enhancer,
        java_common.WARNING,
    ]  # type: List[Stripped]

    code = "\n\n".join(blocks)

    return java_common.JavaFile(
        Stripped("Enhancer.java"),
        f"{code}\n",
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@require(lambda cls: not cls.is_implementation_specific)
def _generate_enhanced_class(
    cls: intermediate.ConcreteClass,
) -> Tuple[Optional[Stripped], Optional[Error]]:
# fmt: on
    """Generate the structure for the enhanced concrete class."""
    enhanced_name = java_naming.class_name(Identifier(f"enhanced_{cls.name}"))
    interface_name = java_naming.interface_name(cls.name)

    blocks = [
        Stripped(f"private final {interface_name} instance;"),
        Stripped(
            f"""\
public {enhanced_name}(
{I}{interface_name} instance,
{I}EnhancementT enhancement
) {{
{I}super(enhancement);
{I}this.instance = instance;
}}"""
        ),
    ]  # type: List[Stripped]

    for prop in cls.properties:
        prop_type = java_common.generate_type(prop.type_annotation)

        inner_type = java_common.generate_type(
            intermediate.beneath_optional(prop.type_annotation)
        )

        prop_name = java_naming.property_name(prop.name)

        getter_name = java_naming.getter_name(prop.name)

        setter_name = java_naming.setter_name(prop.name)

        if isinstance(prop_type, intermediate.OptionalTypeAnnotation):
            blocks.append(
                Stripped(
                    f"""\
@Override
public {prop_type} {getter_name}() {{
{I}return instance.{getter_name}();
}}"""
                )
            )
        else:
            blocks.append(
                Stripped(
                    f"""\
@Override
public {prop_type} {getter_name}() {{
{I}return instance.{getter_name}();
}}"""
                )
            )

        blocks.append(
            Stripped(
                f"""\
@Override
public void {setter_name}({inner_type} {prop_name}) {{
{I}instance.{setter_name}({prop_name});
}}"""
            )
        )

    # region OverXOrEmpty getter

    for prop in cls.properties:
        if isinstance(
            prop.type_annotation, intermediate.OptionalTypeAnnotation
        ) and isinstance(prop.type_annotation.value, intermediate.ListTypeAnnotation):
            prop_name = java_naming.property_name(prop.name)
            method_name = f"over{java_naming.class_name(prop.name)}OrEmpty"
            getter_name = java_naming.getter_name(prop.name)
            items_type = java_common.generate_type(prop.type_annotation.value.items)

            blocks.append(
                Stripped(
                    f"""\
public Iterable<{items_type}> {method_name}() {{
{I}return instance.{method_name}();
}}"""
                )
            )

    # endregion

    for method in cls.methods:
        blocks.append(_generate_delegate_method(method))

    visit_name = java_naming.method_name(Identifier(f"visit_{cls.name}"))

    transform_name = java_naming.method_name(Identifier(f"transform_{cls.name}"))

    blocks.extend(
        [
            Stripped(
                f"""\
public Iterable<IClass> descendOnce() {{
{I}return instance.descendOnce();
}}"""
            ),
            Stripped(
                f"""\
public Iterable<IClass> descend() {{
{I}return instance.descend();
}}"""
            ),
            Stripped(
                f"""\
public void accept(IVisitor visitor) {{
{I}visitor.{visit_name}(instance);
}}"""
            ),
            Stripped(
                f"""\
public <ContextT> void accept(
{I}IVisitorWithContext<ContextT> visitor,
{I}ContextT context
) {{
{I}visitor.{visit_name}(instance, context);
}}"""
            ),
            Stripped(
                f"""\
public <T> T transform(ITransformer<T> transformer) {{
{I}return transformer.{transform_name}(instance);
}}"""
            ),
            Stripped(
                f"""\
public <ContextT, T> T transform(
{I}ITransformerWithContext<ContextT, T> transformer,
{I}ContextT context
) {{
{I}return transformer.{transform_name}(instance, context);
}}"""
            ),
        ]
    )

    writer = io.StringIO()
    writer.write(
        f"""\
public class {enhanced_name}<EnhancementT>
{I}extends Enhanced<EnhancementT>
{I}implements {interface_name} {{
"""
    )

    if len(blocks) > 0:
        for i, block in enumerate(blocks):
            if i > 0:
                writer.write("\n\n")
            writer.write(textwrap.indent(block, I))
    else:
        writer.write("// No properties to wire to instance.")

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
# fmt: on
def _generate_enhanced(
    symbol_table: intermediate.SymbolTable,
    package: java_common.PackageIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[List[java_common.JavaFile]], Optional[List[Error]]]:
    files = []  # type: List[java_common.JavaFile]

    errors = []  # type: List[Error]

    for cls in symbol_table.concrete_classes:
        cls_name = java_naming.class_name(cls.name)

        if cls.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"Enhancing/Enhanced/{cls.name}.java"
            )

            code = spec_impls.get(implementation_key, None)
            if code is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The implementation is missing "
                        f"for the implementation-specific class: {implementation_key}",
                    )
                )
                continue

            assert code is not None

            files.append(
                java_common.JavaFile(
                    f"{cls_name}.java",
                    f"{code}\n",
                ),
            )
        else:
            code, error = _generate_enhanced_class(cls=cls)
            if error is not None:
                errors.append(error)
                continue

            assert code is not None

            imports = [
                Stripped("import java.lang.Iterable;"),
                Stripped("import java.util.Optional;"),
                Stripped("import java.util.List;"),
                Stripped(f"import {package}.visitation.IVisitor;"),
                Stripped(f"import {package}.visitation.IVisitorWithContext;"),
                Stripped(f"import {package}.visitation.ITransformer;"),
                Stripped(f"import {package}.visitation.ITransformerWithContext;"),
                Stripped(f"import {package}.types.enums.*;"),
                Stripped(f"import {package}.types.impl.*;"),
                Stripped(f"import {package}.types.model.*;"),
            ]  # type: List[Stripped]

            blocks = [
                java_common.WARNING,
                Stripped(f"package {package}.enhancing;"),
                Stripped("\n".join(imports)),
                code,
                java_common.WARNING,
            ]  # type: List[Stripped]

            code = Stripped("\n\n".join(blocks))

            files.append(
                java_common.JavaFile(
                    f"Enhanced{cls_name}.java",
                    f"{code}\n",
                ),
            )

    if len(errors) > 0:
        return None, errors

    return files, None


@require(lambda cls: not cls.is_implementation_specific)
def _generate_transform(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the transform method to wrap the instance with an enhancement."""
    blocks = [
        Stripped(
            f"""\
if (that instanceof Enhanced)
{{
{I}throw new IllegalArgumentException(
{II}"The instance has been already enhanced: " + that
{I});
}}"""
        )
    ]  # type: List[Stripped]

    for prop in cls.properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)
        prop_name = java_naming.property_name(prop.name)

        optional = isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)

        wrap_stmt: Stripped

        if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
            # We can not enhance primitive types; nothing to do here.
            continue

        elif isinstance(type_anno, intermediate.OurTypeAnnotation):
            if isinstance(type_anno.our_type, intermediate.Enumeration):
                # We can not enhance enumerations; nothing to do here.
                continue

            elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                # We can not enhance primitive types; nothing to do here.
                continue

            elif isinstance(
                type_anno.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ):
                getter_name = java_naming.getter_name(prop.name)
                setter_name = java_naming.setter_name(prop.name)
                value_interface_name = java_naming.interface_name(
                    type_anno.our_type.name
                )
                transformed_name = java_naming.variable_name(
                    Identifier(f"transformed_{prop.name}")
                )
                casted_name = java_naming.variable_name(
                    Identifier(f"casted_{prop.name}")
                )

                stmt = Stripped(
                    f"""\
IClass {transformed_name} = transform({prop_name});
if (!({transformed_name} instanceof {value_interface_name})) {{
{I}throw new UnsupportedOperationException(
{II}"Expected the transformed value to be a {value_interface_name} " +
{II}", but got: " + {transformed_name}
{I});
}}
{value_interface_name} {casted_name} = ({value_interface_name}) {transformed_name};
that.{setter_name}({casted_name});"""
                )

                writer = io.StringIO()

                if optional:
                    writer.write(
                        f"""\
if (that.{getter_name}().isPresent()) {{
{I}{value_interface_name} {prop_name} = that.{getter_name}().get();
{I}{indent_but_first_line(stmt, I)}
}}"""
                    )
                else:
                    writer.write(
                        f"""\
{value_interface_name} {prop_name} = that.{getter_name}();
{stmt}"""
                    )

                wrap_stmt = Stripped(writer.getvalue())
            else:
                assert_never(type_anno.our_type)
        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            # fmt: off
            assert (
                isinstance(type_anno.items, intermediate.OurTypeAnnotation)
                and isinstance(
                    type_anno.items.our_type,
                    (intermediate.AbstractClass, intermediate.ConcreteClass)
               )
            ), (
                "We handle only lists of classes in the enhancing at the moment. "
                "The meta-model does not contain any other lists, so we wanted to "
                "keep the code as simple as possible, and avoid unrolling. Please "
                "contact the developers if you need this feature."
            )
            # fmt: on

            item_interface_name = java_naming.interface_name(
                type_anno.items.our_type.name
            )
            transformed_name = java_naming.variable_name(
                Identifier(f"transformed_{prop.name}")
            )

            getter_name = java_naming.getter_name(prop.name)

            setter_name = java_naming.setter_name(prop.name)

            stmt = Stripped(
                f"""\
List<{item_interface_name}> {transformed_name} = {prop_name}.stream()
{I}.map(item -> {{
{II}IClass transformed = transform(item);
{II}if (!(transformed instanceof {item_interface_name})) {{
{III}throw new UnsupportedOperationException(
{IIII}"Expected the transformed value to be a {item_interface_name} " +
{IIII}", but got: " + transformed
{III});
{II}}}
{II}return ({item_interface_name}) transformed;
{I}}}).collect(Collectors.toList());
that.{setter_name}({transformed_name});"""
            )

            writer = io.StringIO()

            if optional:
                writer.write(
                    f"""\
if (that.{getter_name}().isPresent()) {{
{I}List<{item_interface_name}> {prop_name} = that.{getter_name}().get();
{I}{indent_but_first_line(stmt, I)}
}}"""
                )
            else:
                writer.write(
                    f"""\
List<{item_interface_name}> {prop_name} = that.{getter_name}();
{stmt}"""
                )

            wrap_stmt = Stripped(writer.getvalue())
        else:
            assert_never(type_anno.our_type)

        blocks.append(wrap_stmt)

    enhanced_name = java_naming.class_name(Identifier(f"enhanced_{cls.name}"))

    blocks.append(
        Stripped(
            f"""\
Optional<EnhancementT> enhancement = enhancementFactory.apply(that);
return !enhancement.isPresent()
{I}? that
{I}: new {enhanced_name}<>(
{II}that,
{II}enhancement.get()
{I});"""
        )
    )

    interface_name = java_naming.interface_name(cls.name)
    transform_name = java_naming.method_name(Identifier(f"transform_{cls.name}"))

    blocks_joined = "\n\n".join(blocks)

    return Stripped(
        f"""\
@Override
public IClass {transform_name}(
{I}{interface_name} that
) {{
{I}{indent_but_first_line(blocks_joined, I)}
}}"""
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_wrapper(
    symbol_table: intermediate.SymbolTable,
    package: java_common.PackageIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[java_common.JavaFile], Optional[List[Error]]]:
    """Generate the transformer that wraps an instance with the enhancement."""
    errors = []  # type: List[Error]
    imports = [
        Stripped("import java.util.List;"),
        Stripped("import java.util.Optional;"),
        Stripped("import java.util.function.Function;"),
        Stripped("import java.util.stream.Collectors;"),
        Stripped("import java.util.stream.Stream;"),
        Stripped(f"import {package}.types.model.*;"),
        Stripped(f"import {package}.visitation.AbstractTransformer;"),
    ]  # type: List[Stripped]
    body = [
        Stripped(
            "private final Function<IClass, Optional<EnhancementT>> "
            "enhancementFactory;"
        ),
        Stripped(
            f"""\
Wrapper(
{I}Function<IClass, Optional<EnhancementT>> enhancementFactory
) {{
{I}this.enhancementFactory = enhancementFactory;
}}"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"Enhancing/Wrap/{cls.name}.java"
            )

            code = spec_impls.get(implementation_key, None)
            if code is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The implementation is missing "
                        f"for the implementation-specific class: {implementation_key}",
                    )
                )
                continue

            body.append(code)
            continue

        body.append(_generate_transform(cls=cls))

    writer = io.StringIO()
    writer.write(
        """\
class Wrapper<EnhancementT> extends AbstractTransformer<IClass> {
"""
    )

    for i, block in enumerate(body):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    blocks = [
        Stripped(java_common.WARNING),
        Stripped(Stripped(f"package {package}.enhancing;")),
        Stripped("\n".join(imports)),
        Stripped(writer.getvalue()),
        Stripped(java_common.WARNING),
    ]  # type: List[Stripped]

    code = Stripped("\n\n".join(blocks))

    return (
        java_common.JavaFile(
            "Wrapper.java",
            f"{code}\n",
        ),
        None,
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    package: java_common.PackageIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[List[java_common.JavaFile]], Optional[List[Error]]]:
    """
    Generate the Java code for enhancing model classes with custom wraps.

    The ``package`` defines the root Java package.
    """

    errors = []  # type: List[Error]

    java_files = [
        _generate_enhanced_abstract_class(package),
        _generate_unwrapper_class(package),
        _generate_enhancer_class(package),
    ]  # type: List[java_common.JavaFile]

    enhanced_files, enhanced_errors = _generate_enhanced(
        symbol_table, package, spec_impls
    )

    if enhanced_errors is not None:
        errors.extend(enhanced_errors)
    else:
        assert enhanced_files is not None

        java_files.extend(enhanced_files)

    wrapper_file, wrapper_errors = _generate_wrapper(symbol_table, package, spec_impls)

    if wrapper_errors is not None:
        errors.extend(wrapper_errors)
    else:
        assert wrapper_file is not None

        java_files.append(wrapper_file)

    if len(errors) > 0:
        return None, errors

    return java_files, None



"""Generate the C# code for shared JSON functionality across unit tests."""

from typing import List

from icontract import ensure

from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.csharp import (
    common as csharp_common,
)
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(namespace: csharp_common.NamespaceIdentifier) -> str:
    """
    Generate the C# code for shared JSON functionality across unit tests.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = [
        Stripped(
            f"""\
public static Nodes.JsonNode ReadFromFile(string path)
{{
{I}using var stream = new FileStream(path, FileMode.Open);
{I}Nodes.JsonNode? node;
{I}try
{I}{{
{II}node = Nodes.JsonNode.Parse(stream);
{I}}}
{I}catch (JsonException exception)
{I}{{
{II}throw new System.InvalidOperationException(
{III}$"Expected the file to be a valid JSON, but it was not: {{path}}; "
{IIII}+ $"exception was: {{exception}}"
{II});
{I}}}

{I}if (node is null)
{I}{{
{II}throw new System.InvalidOperationException(
{III}"Expected the file to be a non-null JSON value, " + $"but it was null: {{path}}"
{II});
{I}}}

{I}return node;
}}"""
        ),
        Stripped(
            f"""\
/// <summary>
/// Serialize <paremref name="something" /> to a uniform JSON text
/// such that we can use it for comparisons in the tests.
/// </summary>
public static Nodes.JsonNode ToJson(object something)
{{
{I}switch (something)
{I}{{
{II}case bool aBool:
{III}return Nodes.JsonValue.Create(aBool);
{II}case long aLong:
{III}return Nodes.JsonValue.Create(aLong);
{II}case double aDouble:
{III}return Nodes.JsonValue.Create(aDouble);
{II}case string aString:
{III}return Nodes.JsonValue.Create(aString)
{IIII}?? throw new System.InvalidOperationException(
{IIIII}$"Could not convert {{something}} " + "to a JSON string"
{IIII});
{II}case byte[] someBytes:
{III}return Nodes.JsonValue.Create(System.Convert.ToBase64String(someBytes))
{IIII}?? throw new System.InvalidOperationException(
{IIIII}$"Could not convert {{something}} to " + "a base64-encoded JSON string"
{IIII});
{II}case Aas.IClass instance:
{III}return Aas.Jsonization.Serialize.ToJsonObject(instance);
{II}default:
{III}throw new System.ArgumentException(
{IIII}$"The conversion of type {{something.GetType()}} "
{IIIII}+ $"to a JSON node has not been defined: {{something}}"
{III});
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/// <summary>
/// Infer the node kind of the JSON node.
/// </summary>
/// <remarks>
/// This function is necessary since NET6 does not fully support node kinds yet.
/// See:
/// <ul>
/// <li>https://github.com/dotnet/runtime/issues/53406</li>
/// <li>https://github.com/dotnet/runtime/issues/55827</li>
/// <li>https://github.com/dotnet/runtime/issues/56592</li>
/// </ul>
/// </remarks>
private static string GetNodeKind(Nodes.JsonNode node)
{{
{I}switch (node)
{I}{{
{II}case Nodes.JsonArray _:
{III}return "array";
{II}case Nodes.JsonObject _:
{III}return "object";
{II}case Nodes.JsonValue _:
{III}return "value";
{II}default:
{III}throw new System.InvalidOperationException(
{IIII}$"Unhandled JsonNode: {{node.GetType()}}"
{III});
{I}}}
}}"""
        ),
        Stripped(
            f"""\
public static void CheckJsonNodesEqual(
{I}Nodes.JsonNode that,
{I}Nodes.JsonNode other,
{I}out Reporting.Error? error
)
{{
{I}error = null;

{I}var thatNodeKind = GetNodeKind(that);
{I}var otherNodeKind = GetNodeKind(other);

{I}if (thatNodeKind != otherNodeKind)
{I}{{
{II}error = new Reporting.Error(
{III}$"Mismatch in node kinds : {{thatNodeKind}} != {{otherNodeKind}}"
{II});
{II}return;
{I}}}

{I}switch (that)
{I}{{
{II}case Nodes.JsonArray thatArray:
{III}{{
{IIII}var otherArray = (other as Nodes.JsonArray)!;
{IIII}if (thatArray.Count != otherArray.Count)
{IIII}{{
{IIIII}error = new Reporting.Error(
{IIIII}    $"Unequal array lengths: {{thatArray.Count}} != {{otherArray.Count}}"
{IIIII});
{IIIII}return;
{IIII}}}

{IIII}for (int i = 0; i < thatArray.Count; i++)
{IIII}{{
{IIIII}CheckJsonNodesEqual(thatArray[i]!, otherArray[i]!, out error);
{IIIII}if (error != null)
{IIIII}{{
{IIIII}    error.PrependSegment(new Reporting.IndexSegment(i));
{IIIII}    return;
{IIIII}}}
{IIII}}}

{IIII}break;
{III}}}
{II}case Nodes.JsonObject thatObject:
{III}{{
{IIII}var thatDictionary = thatObject as IDictionary<string, Nodes.JsonNode>;
{IIII}var otherDictionary = (other as IDictionary<string, Nodes.JsonNode>)!;

{IIII}var thatKeys = thatDictionary.Keys.ToList();
{IIII}thatKeys.Sort();

{IIII}var otherKeys = otherDictionary.Keys.ToList();
{IIII}otherKeys.Sort();

{IIII}if (!thatKeys.SequenceEqual(otherKeys))
{IIII}{{
{IIIII}error = new Reporting.Error(
{IIIII}    "Objects with different properties: "
{IIIII}        + $"{{string.Join(", ", thatKeys)}} != "
{IIIII}        + $"{{string.Join(", ", otherKeys)}}"
{IIIII});
{IIIII}return;
{IIII}}}

{IIII}foreach (var key in thatKeys)
{IIII}{{
{IIIII}CheckJsonNodesEqual(thatDictionary[key], otherDictionary[key], out error);
{IIIII}if (error != null)
{IIIII}{{
{IIIII}    error.PrependSegment(new Reporting.NameSegment(key));
{IIIII}    return;
{IIIII}}}
{IIII}}}

{IIII}break;
{III}}}
{II}case Nodes.JsonValue thatValue:
{III}{{
{IIII}string thatAsJsonString = thatValue.ToJsonString();

{IIII}// NOTE (mristin):
{IIII}// This is slow, but there is no way around it at the moment with NET6.
{IIII}// See:
{IIII}// * https://github.com/dotnet/runtime/issues/56592
{IIII}// * https://github.com/dotnet/runtime/issues/55827
{IIII}// * https://github.com/dotnet/runtime/issues/53406
{IIII}var otherValue = (other as Nodes.JsonValue)!;
{IIII}string otherAsJsonString = otherValue.ToJsonString();

{IIII}if (thatAsJsonString != otherAsJsonString)
{IIII}{{
{IIIII}error = new Reporting.Error(
{IIIII}    $"Unequal values: {{thatAsJsonString}} != {{otherAsJsonString}}"
{IIIII});
{IIIII}// ReSharper disable once RedundantJumpStatement
{IIIII}return;
{IIII}}}

{IIII}break;
{III}}}
{II}default:
{III}throw new System.InvalidOperationException(
{IIII}$"Unhandled JSON node: {{that.GetType()}}"
{III});
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}
    
using Aas = {namespace}; // renamed

using FileMode = System.IO.FileMode;
using FileStream = System.IO.FileStream;
using JsonException = System.Text.Json.JsonException;
using Nodes = System.Text.Json.Nodes;

using System.Collections.Generic; // can't alias
using System.Linq; // can't alias

namespace {namespace}.Tests
{{
{I}public static class CommonJson
{I}{{
{I}{indent_but_first_line(blocks_joined, I)}
{I}}}  // public static class CommonJson
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())
